#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import functools
import json
from typing import Optional, Any
from dataclasses import dataclass, field, fields
from urllib.parse import urlparse

import yaml

from splunk_add_on_ucc_framework import utils
from splunk_add_on_ucc_framework import app_manifest as app_manifest_lib
from splunk_add_on_ucc_framework.commands.rest_builder.user_defined_rest_handlers import (
    UserDefinedRestHandlers,
)
from splunk_add_on_ucc_framework.entity import expand_entity
from splunk_add_on_ucc_framework.tabs import resolve_tab, LoggingTab

Loader = getattr(yaml, "CSafeLoader", yaml.SafeLoader)
yaml_load = functools.partial(yaml.load, Loader=Loader)


@dataclass(frozen=True)
class OSDependentLibraryConfig:
    name: str
    version: str
    platform: str
    python_version: str
    target: str
    os: str
    deps_flag: str
    dependencies: bool = field(default=False)
    ignore_requires_python: bool = field(default=False)

    @classmethod
    def from_dict(cls, **kwargs: Any) -> "OSDependentLibraryConfig":
        result = {
            dc_field.name: kwargs[dc_field.name]
            for dc_field in fields(cls)
            if dc_field.name in kwargs
            and (
                isinstance(kwargs[dc_field.name], dc_field.type)
                or kwargs[dc_field.name] is None
            )
        }
        deps_flag = "" if result.get("dependencies") else "--no-deps"
        ignore_python = (
            "--ignore-requires-python" if result.get("ignore_requires_python") else ""
        )
        result.update({"deps_flag": deps_flag})
        result.update({"ignore_requires_python": ignore_python})
        result.update({"python_version": result["python_version"]})
        return cls(**result)


class GlobalConfig:
    def __init__(
        self,
        content: dict[str, Any],
        is_yaml: bool,
    ) -> None:
        self._content = content
        self._is_global_config_yaml = is_yaml
        rest_handlers = self._content.get("options", {}).get("restHandlers", [])
        self.user_defined_handlers = UserDefinedRestHandlers(rest_handlers)

    @classmethod
    def from_file(cls, global_config_path: str) -> "GlobalConfig":
        with open(global_config_path) as f_config:
            config_raw = f_config.read()
        is_global_config_yaml = True if global_config_path.endswith(".yaml") else False
        content = (
            yaml_load(config_raw) if is_global_config_yaml else json.loads(config_raw)
        )
        return GlobalConfig(content, is_global_config_yaml)

    @classmethod
    def from_app_manifest(
        cls, app_manifest: app_manifest_lib.AppManifest
    ) -> "GlobalConfig":
        content = {
            "meta": {
                "name": app_manifest.get_addon_name(),
                # TODO(ADDON-79208): once `restRoot` is optional, this line can be removed
                "restRoot": app_manifest.get_addon_name(),
                "displayName": app_manifest.get_title(),
                "version": app_manifest.get_addon_version(),
            }
        }
        return GlobalConfig(
            content,
            False,
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, GlobalConfig):
            raise NotImplementedError()
        return all(
            [
                self.content == other.content,
                self.is_yaml == other.is_yaml,
            ]
        )

    def dump(self, path: str) -> None:
        if self.is_yaml:
            utils.dump_yaml_config(self.content, path)
        else:
            utils.dump_json_config(self.content, path)

    def expand(self) -> None:
        self.expand_tabs()
        self.expand_entities()

    def expand_tabs(self) -> None:
        if self.has_pages() and self.has_configuration():
            for i, tab in enumerate(self._content["pages"]["configuration"]["tabs"]):
                self._content["pages"]["configuration"]["tabs"][i] = resolve_tab(tab)

    def expand_entities(self) -> None:
        self._expand_entities(
            self._content.get("pages", {}).get("configuration", {}).get("tabs")
        )
        self._expand_entities(
            self._content.get("pages", {}).get("inputs", {}).get("services")
        )

    @staticmethod
    def _expand_entities(items: Optional[list[dict[Any, Any]]]) -> None:
        if items is None:
            return

        for item in items:
            for i, entity in enumerate(item.get("entity", [])):
                item["entity"][i] = expand_entity(entity)

    @property
    def content(self) -> Any:
        return self._content

    @property
    def is_yaml(self) -> bool:
        return self._is_global_config_yaml

    @property
    def inputs(self) -> list[Any]:
        if self.has_pages() and "inputs" in self._content["pages"]:
            return self._content["pages"]["inputs"]["services"]
        return []

    @property
    def configuration(self) -> list[Any]:
        if self.has_pages() and "configuration" in self._content["pages"]:
            return self._content["pages"]["configuration"]["tabs"]
        return []

    @property
    def endpointUrl(self) -> list[str]:
        def extract_urls(entities_list: list[Any]) -> set[str]:
            urls = set()
            for content in entities_list:
                for entity in content.get("entity", []):
                    if entity.get("type") in ("multipleSelect", "singleSelect"):
                        endpoint = entity.get("options", {}).get("endpointUrl")
                        # only include those endpoint which starts with data
                        # the other ones are covered if they are defined in options->
                        # TODO; If add-ons uses any other endpoints apart from data then add it here.
                        if endpoint and endpoint.startswith("data"):
                            urls.add(urlparse(endpoint).path)
            return urls

        urls = set()
        if self.has_configuration():
            urls.update(extract_urls(self.configuration))

        if self.has_inputs():
            urls.update(extract_urls(self.inputs))

        return list(urls)

    @property
    def pages(self) -> list[Any]:
        if "pages" in self._content:
            return self._content["pages"]
        return []

    @property
    def resolved_configuration(self) -> list[Any]:
        if self.has_pages() and "configuration" in self.pages:
            return [resolve_tab(tab) for tab in self.configuration]
        return []

    @property
    def dashboard(self) -> dict[str, Any]:
        return self._content.get("pages", {}).get("dashboard")

    @property
    def settings(self) -> list[Any]:
        settings = []
        for tab in self.configuration:
            if "table" not in tab:
                settings.append(tab)
        return settings

    @property
    def logging_tab(self) -> dict[str, Any]:
        for tab in self.configuration:
            if LoggingTab.from_definition(tab) is not None:
                return tab
        return {}

    @property
    def configs(self) -> list[Any]:
        configs = []
        for tab in self.configuration:
            if "table" in tab:
                configs.append(tab)
        return configs

    @property
    def alerts(self) -> list[dict[str, Any]]:
        return self._content.get("alerts", [])

    @property
    def custom_search_commands(self) -> list[dict[str, Any]]:
        return self._content.get("customSearchCommand", [])

    @property
    def meta(self) -> dict[str, Any]:
        return self._content["meta"]

    @property
    def namespace(self) -> str:
        return self.meta["restRoot"]

    @property
    def product(self) -> str:
        return self.meta["name"]

    @property
    def display_name(self) -> str:
        return self.meta["displayName"]

    @property
    def version(self) -> str:
        return self.meta["version"]

    def capabilities(self, **kwargs: bool) -> dict[str, str]:
        if kwargs.get("config") and self.has_configuration():
            return self._content["pages"]["configuration"].get("capabilities", {})
        elif kwargs.get("inputs") and self.has_inputs():
            return self._content["pages"]["inputs"].get("capabilities", {})
        return {}

    @property
    def ucc_version(self) -> str:
        return self.meta["_uccVersion"]

    @property
    def schema_version(self) -> Optional[str]:
        return self.meta.get("schemaVersion")

    @property
    def os_libraries(self) -> Optional[list[OSDependentLibraryConfig]]:
        if self._content["meta"].get("os-dependentLibraries"):
            return [
                OSDependentLibraryConfig.from_dict(**lib)
                for lib in self._content["meta"].get("os-dependentLibraries")
            ]
        return None

    def update_schema_version(self, new_schema_version: str) -> None:
        self.meta["schemaVersion"] = new_schema_version

    def update_addon_version(self, version: str) -> None:
        self._content.setdefault("meta", {})["version"] = version

    def cleanup_unwanted_params(self) -> None:
        if "_uccVersion" in self.content["meta"]:
            del self.content["meta"]["_uccVersion"]

    def add_ucc_version(self, version: str) -> None:
        self.content.setdefault("meta", {})["_uccVersion"] = version

    def has_pages(self) -> bool:
        return bool(self.pages)

    def has_inputs(self) -> bool:
        return bool(self.inputs)

    def has_configuration(self) -> bool:
        return bool(self.configuration)

    def has_alerts(self) -> bool:
        return bool(self.alerts)

    def has_custom_search_commands(self) -> bool:
        return bool(self.custom_search_commands)

    def has_dashboard(self) -> bool:
        return self.dashboard is not None

    def has_oauth(self) -> bool:
        for tab in self.configuration:
            if tab["name"] == "account":
                for entity in tab["entity"]:
                    if entity["type"] == "oauth":
                        return True
        return False



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Union, Optional

from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.base import (
    RestEndpointBuilder,
)
from splunk_add_on_ucc_framework.commands.rest_builder.user_defined_rest_handlers import (
    EndpointRegistrationEntry,
)
from splunk_add_on_ucc_framework.generators.file_generator import FileGenerator
from splunk_add_on_ucc_framework.global_config import GlobalConfig


class RestMapConf(FileGenerator):
    __description__ = (
        "Generates `restmap.conf` for the custom REST handlers that "
        "are generated based on configs from globalConfig"
    )

    def __init__(
        self, global_config: GlobalConfig, input_dir: str, output_dir: str
    ) -> None:
        super().__init__(global_config, input_dir, output_dir)
        self.conf_file = "restmap.conf"
        self.endpoints: list[Union[RestEndpointBuilder, EndpointRegistrationEntry]] = []
        self.configuration_endpoints: list[
            Union[RestEndpointBuilder, EndpointRegistrationEntry]
        ] = []
        self.inputs_endpoints: list[
            Union[RestEndpointBuilder, EndpointRegistrationEntry]
        ] = []
        self.custom_endpoints: list[
            Union[RestEndpointBuilder, EndpointRegistrationEntry]
        ] = []

        if global_config.has_pages():
            self.configuration_endpoints.extend(self._gc_schema.configuration_endpoints)
            self.inputs_endpoints.extend(self._gc_schema.inputs_endpoints)
            self.namespace = self._gc_schema.namespace
        self.custom_endpoints.extend(
            global_config.user_defined_handlers.endpoint_registration_entries
        )

        self.configuration_endpoint_names = ", ".join(
            sorted([ep.name for ep in self.configuration_endpoints])
        )
        self.inputs_endpoint_names = ", ".join(
            sorted([ep.name for ep in self.inputs_endpoints])
        )
        self.endpoints.extend(self.configuration_endpoints)
        self.endpoints.extend(self.inputs_endpoints)
        self.endpoints.extend(self.custom_endpoints)

    def generate(self) -> Optional[list[dict[str, str]]]:
        if not self.endpoints:
            return None

        file_path = self.get_file_output_path(["default", self.conf_file])
        self.set_template_and_render(
            template_file_path=["conf_files"], file_name="restmap_conf.template"
        )
        rendered_content = self._template.render(
            endpoints=self.endpoints,
            configuration_endpoints=self.configuration_endpoints,
            inputs_endpoints=self.inputs_endpoints,
            configuration_endpoint_names=self.configuration_endpoint_names,
            inputs_endpoint_names=self.inputs_endpoint_names,
            namespace=self.namespace,
            configuration_capability=self._global_config.capabilities(config=True),
            input_capability=self._global_config.capabilities(inputs=True),
            custom_endpoints=self.custom_endpoints,
        )
        return [
            {
                "file_name": self.conf_file,
                "file_path": file_path,
                "content": rendered_content,
            }
        ]



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from splunk_add_on_ucc_framework.generators.file_generator import FileGenerator
from typing import Optional
import os
from xml.etree.ElementTree import Element, SubElement, tostring
from splunk_add_on_ucc_framework.utils import pretty_print_xml
from splunk_add_on_ucc_framework.global_config import GlobalConfig
import logging

logger = logging.getLogger("ucc_gen")


class DefaultXml(FileGenerator):
    __description__ = (
        "Generates default.xml file based on configs present in globalConfig,"
        " in `default/data/ui/nav` folder."
    )

    def __init__(
        self, global_config: GlobalConfig, input_dir: str, output_dir: str
    ) -> None:
        super().__init__(global_config, input_dir, output_dir)
        addon_name = self._addon_name
        if not isinstance(addon_name, str):
            raise ValueError("addon_name must be a string")
        default_ui_path = os.path.join(
            self._output_dir, addon_name, "default", "data", "ui"
        )
        default_xml_path = os.path.join(default_ui_path, "nav", "default.xml")
        if os.path.exists(default_xml_path):
            logger.info(
                "Skipping generating data/ui/nav/default.xml because file already exists."
            )
        else:
            if global_config.has_pages():
                self.default_xml_content = self.generate_nav_default_xml(
                    include_inputs=global_config.has_inputs(),
                    include_dashboard=global_config.has_dashboard(),
                    include_configuration=global_config.has_configuration(),
                    default_view=global_config.meta.get("defaultView"),
                )

    def generate_nav_default_xml(
        self,
        include_inputs: bool,
        include_dashboard: bool,
        include_configuration: bool,
        default_view: Optional[str],
    ) -> str:
        """
        Generates `default/data/ui/nav/default.xml` file.

        The validation is being done in `_validate_meta_default_view` function from `global_config_validator.py` file.
        """
        nav = Element("nav")
        if default_view is None:
            # we do this calculation as all the below properties are now optional
            if include_configuration:
                default_view = "configuration"
            elif include_inputs:
                default_view = "inputs"
            elif include_dashboard:
                default_view = "dashboard"
            else:
                default_view = "search"

        if include_inputs:
            if default_view == "inputs":
                SubElement(nav, "view", attrib={"name": "inputs", "default": "true"})
            else:
                SubElement(nav, "view", attrib={"name": "inputs"})

        if include_configuration:
            if default_view == "configuration":
                SubElement(
                    nav, "view", attrib={"name": "configuration", "default": "true"}
                )
            else:
                SubElement(nav, "view", attrib={"name": "configuration"})

        if include_dashboard:
            if default_view == "dashboard":
                SubElement(nav, "view", attrib={"name": "dashboard", "default": "true"})
            else:
                SubElement(nav, "view", attrib={"name": "dashboard"})

        if default_view == "search":
            SubElement(nav, "view", attrib={"name": "search", "default": "true"})
        else:
            SubElement(nav, "view", attrib={"name": "search"})

        nav_as_string = tostring(nav, encoding="unicode")
        return pretty_print_xml(nav_as_string)

    def generate(self) -> Optional[list[dict[str, str]]]:
        if not self._global_config.has_pages():
            return None
        file_path = self.get_file_output_path(
            ["default", "data", "ui", "nav", "default.xml"]
        )
        return [
            {
                "file_name": "default.xml",
                "file_path": file_path,
                "content": self.default_xml_content,
            }
        ]



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from os.path import isfile, join
from typing import Optional
from splunk_add_on_ucc_framework.generators.file_generator import FileGenerator
from splunk_add_on_ucc_framework.global_config import GlobalConfig


class ServerConf(FileGenerator):
    __description__ = (
        "Generates `server.conf` for the custom conf "
        "files created as per configurations in globalConfig"
    )

    def __init__(
        self, global_config: GlobalConfig, input_dir: str, output_dir: str
    ) -> None:
        super().__init__(global_config, input_dir, output_dir)
        self.conf_file = "server.conf"
        self.custom_conf = []
        self.custom_conf.extend(list(self._gc_schema.settings_conf_file_names))
        self.custom_conf.extend(list(self._gc_schema.configs_conf_file_names))
        self.custom_conf.extend(list(self._gc_schema.oauth_conf_file_names))

    def generate(self) -> Optional[list[dict[str, str]]]:
        if not self.custom_conf:
            return None

        file_path = self.get_file_output_path(["default", self.conf_file])
        # For now, only create server.conf only if
        # no server.conf is present in the source package.
        if isfile(join(self._input_dir, "default", self.conf_file)):
            return None
        self.set_template_and_render(
            template_file_path=["conf_files"], file_name="server_conf.template"
        )
        rendered_content = self._template.render(custom_conf=self.custom_conf)
        return [
            {
                "file_name": self.conf_file,
                "file_path": file_path,
                "content": rendered_content,
            }
        ]



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import ast
from pathlib import Path
from typing import Any, Optional
from splunk_add_on_ucc_framework.generators.file_generator import FileGenerator
from splunk_add_on_ucc_framework.global_config import GlobalConfig


class CustomCommandPy(FileGenerator):
    __description__ = "Generates Python files for custom search commands provided in the globalConfig."

    def __init__(
        self, global_config: GlobalConfig, input_dir: str, output_dir: str
    ) -> None:
        super().__init__(global_config, input_dir, output_dir)
        self.commands_info = []
        for command in global_config.custom_search_commands:
            argument_list: list[str] = []
            import_map = False
            imported_file_name = command["fileName"].replace(".py", "")
            template = command["commandType"].replace(" ", "_") + ".template"
            if command["commandType"] == "transforming":
                module_path = Path(
                    os.path.realpath(
                        os.path.join(self._input_dir, "bin", command["fileName"])
                    )
                )

                if not module_path.is_file():
                    raise FileNotFoundError(
                        f"Module path '{module_path}' does not point to a valid file."
                    )
                module_content = ast.parse(module_path.read_text(encoding="utf-8"))
                for node in module_content.body:
                    if isinstance(node, ast.FunctionDef) and node.name == "map":
                        import_map = True

            for argument in command["arguments"]:
                argument_dict = {
                    "name": argument["name"],
                    "require": argument.get("required", False),
                    "validate": argument.get("validate"),
                    "default": argument.get("defaultValue"),
                }
                self.argument_generator(argument_list, argument_dict)
            self.commands_info.append(
                {
                    "imported_file_name": imported_file_name,
                    "file_name": command["commandName"],
                    "class_name": command["commandName"].title(),
                    "description": command.get("description"),
                    "syntax": command.get("syntax"),
                    "template": template,
                    "list_arg": argument_list,
                    "import_map": import_map,
                }
            )

    def argument_generator(
        self, argument_list: list[str], arg: dict[str, Any]
    ) -> list[str]:
        validate_str = ""
        validate = arg.get("validate", {})
        if validate:
            validate_type = validate["type"]
            if validate_type in ("Integer", "Float"):
                min_val = validate.get("minimum")
                max_val = validate.get("maximum")
                args = []
                if min_val is not None:
                    args.append(f"minimum={min_val}")
                if max_val is not None:
                    args.append(f"maximum={max_val}")
                validate_args = ", ".join(args)
                validate_str = (
                    f", validate=validators.{validate_type}({validate_args})"
                    if args
                    else f", validate=validators.{validate_type}()"
                )
            else:
                validate_str = f", validate=validators.{validate_type}()"

        if arg["default"] is None:
            arg_str = (
                f"{arg['name']} = Option(name='{arg['name']}', "
                f"require={arg.get('require')}"
                f"{validate_str})"
            )
        else:
            arg_str = (
                f"{arg['name']} = Option(name='{arg['name']}', "
                f"require={arg.get('require')}"
                f"{validate_str}, "
                f"default='{arg.get('default', '')}')"
            )
        argument_list.append(arg_str)
        return argument_list

    def generate(self) -> Optional[list[dict[str, str]]]:
        if not self.commands_info:
            return None

        generated_files = []
        for command_info in self.commands_info:
            file_name = command_info["file_name"] + ".py"
            file_path = self.get_file_output_path(["bin", file_name])
            self.set_template_and_render(
                template_file_path=["custom_command"],
                file_name=command_info["template"],
            )
            rendered_content = self._template.render(
                imported_file_name=command_info["imported_file_name"],
                class_name=command_info["class_name"],
                description=command_info["description"],
                syntax=command_info["syntax"],
                list_arg=command_info["list_arg"],
                import_map=command_info["import_map"],
            )
            generated_files.append(
                {
                    "file_name": file_name,
                    "file_path": file_path,
                    "content": rendered_content,
                }
            )
        return generated_files



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional

from splunk_add_on_ucc_framework.generators.file_generator import FileGenerator
from splunk_add_on_ucc_framework.global_config import GlobalConfig


class SettingsConf(FileGenerator):
    __description__ = (
        "Generates `<YOUR_ADDON_NAME>_settings.conf.spec` "
        "file for the Proxy, Logging or Custom Tab mentioned in globalConfig"
    )

    def __init__(
        self, global_config: GlobalConfig, input_dir: str, output_dir: str
    ) -> None:
        super().__init__(global_config, input_dir, output_dir)
        self.settings_stanzas: list[tuple[str, list[str]]] = []
        self.default_content: str = ""

        if global_config.has_configuration():
            self.conf_file = global_config.namespace.lower() + "_settings.conf"
            self.conf_spec_file = f"{self.conf_file}.spec"
            for setting in global_config.settings:
                content = self._gc_schema._get_oauth_enitities(
                    setting.get("entity", [])
                )
                fields, _ = self._gc_schema._parse_fields(content)
                self.settings_stanzas.append(
                    (setting["name"], [f"{f._name} = " for f in fields])
                )
                if setting.get("entity") is None:
                    # When a tab doesn't have an entity, we add newlines as
                    # done in `generate_conf_with_default_values()` function.
                    self.default_content = f"[{setting['name']}]\n\n"
            if self._gc_schema._endpoints.get("settings") is not None:
                self.default_content += self._gc_schema._endpoints[
                    "settings"
                ].generate_conf_with_default_values()

    def generate(self) -> Optional[list[dict[str, str]]]:
        conf_files: list[dict[str, str]] = []
        conf = self.generate_conf()
        conf_spec = self.generate_conf_spec()
        if conf is not None:
            conf_files.append(conf)
        if conf_spec is not None:
            conf_files.append(conf_spec)
        return None if conf_files == [] else conf_files

    def generate_conf(self) -> Optional[dict[str, str]]:
        if not self.default_content:
            return None
        file_path = self.get_file_output_path(["default", self.conf_file])
        self.set_template_and_render(
            template_file_path=["conf_files"], file_name="settings_conf.template"
        )

        rendered_content = self._template.render(default_content=self.default_content)
        return {
            "file_name": self.conf_file,
            "file_path": file_path,
            "content": rendered_content,
        }

    def generate_conf_spec(self) -> Optional[dict[str, str]]:
        if not self.settings_stanzas:
            return None

        file_path = self.get_file_output_path(["README", self.conf_spec_file])
        self.set_template_and_render(
            template_file_path=["README"], file_name="settings_conf_spec.template"
        )

        rendered_content = self._template.render(settings_stanzas=self.settings_stanzas)
        return {
            "file_name": self.conf_spec_file,
            "file_path": file_path,
            "content": rendered_content,
        }



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from os.path import dirname, join, realpath

from splunk_add_on_ucc_framework.generators import GEN_FILE_LIST


def generate_docs() -> str:
    content = []
    content.append(
        """---
title: UCC framework generated files
---

The following table describes the files generated by UCC framework.
"""
    )
    content.append("| File Name  | File Location | File Description |")
    content.append("| ------------ | ------------ | ----------------- |")
    for file_name, file_class, file_path in GEN_FILE_LIST:
        content.append(
            f"| {file_name} | output/&lt;YOUR_ADDON_NAME&gt;/{'/'.join(file_path)} "
            f"| {file_class.__description__} |"
        )
    # statically appending this line as we are generating it now
    content.append(
        "| globalConfig.json | &lt;source_dir&gt; | "
        "Generates globalConfig.json file in the source code if globalConfig "
        "is not present in source directory at build time. |"
    )
    content.append("\n")
    return "\n".join(content)


def main() -> None:
    content = generate_docs()

    with open(
        join(
            realpath(dirname(dirname(dirname(__file__)))), "docs", "generated_files.md"
        ),
        "w",
    ) as writer:
        writer.write(content)


if __name__ == "__main__":
    main()



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Union, Optional

from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.base import (
    RestEndpointBuilder,
)
from splunk_add_on_ucc_framework.commands.rest_builder.user_defined_rest_handlers import (
    EndpointRegistrationEntry,
)
from splunk_add_on_ucc_framework.generators.file_generator import FileGenerator
from splunk_add_on_ucc_framework.global_config import GlobalConfig

actions_to_methods = {
    "edit": ["POST", "GET"],
    "remove": ["DELETE"],
    "list": ["GET"],
    "create": ["POST"],
}


class WebConf(FileGenerator):
    __description__ = (
        "Generates `web.conf` to expose the endpoints generated in "
        "`restmap.conf` which is generated based on configurations from globalConfig."
    )

    def __init__(
        self, global_config: GlobalConfig, input_dir: str, output_dir: str
    ) -> None:
        super().__init__(global_config, input_dir, output_dir)

    def get_methods_for_actions(self, actions: list[str]) -> list[str]:
        methods = []
        for action in actions:
            methods.extend(actions_to_methods.get(action, []))

        return sorted(list(set(methods)))

    def generate(self) -> Optional[list[dict[str, str]]]:
        if not self._global_config.has_pages():
            return None

        endpoints: list[Union[RestEndpointBuilder, EndpointRegistrationEntry]] = []
        custom_endpoints: list[
            Union[RestEndpointBuilder, EndpointRegistrationEntry]
        ] = []
        endpoints.extend(self._gc_schema.endpoints)
        methods = {}
        custom_endpoints.extend(
            self._global_config.user_defined_handlers.endpoint_registration_entries
        )

        for endpoint in custom_endpoints:
            methods[endpoint.name] = self.get_methods_for_actions(endpoint.actions())
        conf_file = "web.conf"

        file_path = self.get_file_output_path(["default", conf_file])
        rendered_content = self._render(
            "web_conf.template",
            endpoints=endpoints,
            custom_endpoints=custom_endpoints,
            methods=methods,
            endpointUrl=self._global_config.endpointUrl,
        )
        return [
            {
                "file_name": conf_file,
                "file_path": file_path,
                "content": rendered_content,
            }
        ]



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from time import time

from splunk_add_on_ucc_framework.generators.file_generator import FileGenerator
from splunk_add_on_ucc_framework.global_config import GlobalConfig
from splunk_add_on_ucc_framework.utils import get_app_manifest


class AppConf(FileGenerator):
    __description__ = (
        "Generates `app.conf` with the details mentioned in globalConfig[meta]"
    )

    def __init__(self, global_config: GlobalConfig, input_dir: str, output_dir: str):
        super().__init__(global_config, input_dir, output_dir)
        app_manifest = get_app_manifest(input_dir)
        self.description = app_manifest.get_description()
        self.title = app_manifest.get_title()
        self.author = app_manifest.get_authors()[0]["name"]
        self.conf_file = "app.conf"
        self.check_for_updates = "true"
        self.custom_conf = []
        self.name = self._addon_name
        self.id = self._addon_name
        self.supported_themes = ""

        self.custom_conf.extend(list(self._gc_schema.settings_conf_file_names))
        self.custom_conf.extend(list(self._gc_schema.configs_conf_file_names))
        self.custom_conf.extend(list(self._gc_schema.oauth_conf_file_names))

        if global_config.meta.get("checkForUpdates") is False:
            self.check_for_updates = "false"
        if global_config.meta.get("supportedThemes") is not None:
            self.supported_themes = ", ".join(global_config.meta["supportedThemes"])

        self.addon_version = global_config.version
        self.is_visible = str(
            global_config.meta.get("isVisible", global_config.has_pages())
        ).lower()
        self.build = str(int(time()))

    def generate(self) -> list[dict[str, str]]:
        file_path = self.get_file_output_path(["default", self.conf_file])
        self.set_template_and_render(
            template_file_path=["conf_files"], file_name="app_conf.template"
        )
        rendered_content = self._template.render(
            custom_conf=self.custom_conf,
            addon_version=self.addon_version,
            check_for_updates=self.check_for_updates,
            supported_themes=self.supported_themes,
            description=self.description,
            author=self.author,
            name=self.name,
            build=self.build,
            id=self.id,
            label=self.title,
            is_visible=self.is_visible,
        )
        return [
            {
                "file_name": self.conf_file,
                "file_path": file_path,
                "content": rendered_content,
                "merge_mode": "item_overwrite",
            }
        ]



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Any

from splunk_add_on_ucc_framework.tabs.proxy_tab import ProxyTab
from splunk_add_on_ucc_framework.tabs.logging_tab import LoggingTab
from splunk_add_on_ucc_framework.tabs.tab import Tab


TAB_TYPES: list[type[Tab]] = [LoggingTab, ProxyTab]


def resolve_tab(tab_definition: dict[Any, Any]) -> Tab:
    """
    Convert the tab dictionary into a tab object. It tries to initialize every type from TAB_TYPES. If there are
    no matches, it initializes Tab class.

    Args:
        tab_definition: Tab definition. It can be a dictionary in the classic form, e.g.:
            {
                "name": "...",
                "title": "...",
                "entity": [...]
            }

            or a special component, e.g.:
            { "type": "loggingTab" }

    Returns: Tab instance or its subclass

    """
    for tab_type in TAB_TYPES:
        tab_obj = tab_type.from_definition(tab_definition)

        if tab_obj is not None:
            return tab_obj

    return Tab(tab_definition)



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from string import Template
from typing import Optional, TYPE_CHECKING, Any

from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.base import (
    RestEndpointBuilder,
    RestEntityBuilder,
)

if TYPE_CHECKING:
    from splunk_add_on_ucc_framework.commands.rest_builder.endpoint.field import (
        RestFieldBuilder,
    )


class SingleModelEntityBuilder(RestEntityBuilder):
    def __init__(
        self, name: Optional[str], fields: list["RestFieldBuilder"], **kwargs: Any
    ) -> None:
        super().__init__(name, fields, **kwargs)

    @property
    def name_spec(self) -> str:
        return "<name>"

    @property
    def name_rh(self) -> str:
        return ""


class SingleModelEndpointBuilder(RestEndpointBuilder):
    _rh_template = """
import import_declare_test

from splunktaucclib.rest_handler.endpoint import (
    field,
    validator,
    RestModel,
    SingleModel,
)
from splunktaucclib.rest_handler import admin_external, util
from {handler_module} import {handler_class}
import logging
{additional_imports}

util.remove_http_proxy_env_vars()

{entity}

endpoint = SingleModel(
    '{conf_name}',
    model,
    config_name='{config_name}',
    need_reload={need_reload},
)
{additional_code}

if __name__ == '__main__':
    logging.getLogger().addHandler(logging.NullHandler())
    admin_external.handle(
        endpoint,
        handler={handler_class_used},
    )
"""

    def actions(self) -> list[str]:
        return ["edit", "list", "remove", "create"]

    def generate_rh(self) -> str:
        entity = self._entities[0]
        return self._rh_template.format(
            handler_module=self.rh_module,
            handler_class=self.rh_class,
            handler_class_used=self.rh_class,
            entity=entity.generate_rh(),
            conf_name=self.conf_name,
            config_name=self._name,
            need_reload=self.need_reload,
            additional_code="",
            additional_imports="",
        )


class SingleModelEndpointBuilderWithOauth(SingleModelEndpointBuilder):
    _imports = "".join(
        (
            "import json\n",
            "from solnlib import splunk_rest_client as rest_client\n",
            "from splunk.admin import InternalException\n",
        )
    )
    _cls_template = """
APP_NAME = ${app_name}
OAUTH_ENDPOINT = ${oauth_endpoint}
TOKEN_ENDPOINT = ${token_endpoint}


class ${class_name}(${base_class}):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._oauth_url = f"/servicesNS/nobody/{APP_NAME}/{OAUTH_ENDPOINT}/oauth"
        self._rest_client = rest_client.SplunkRestClient(
            self.getSessionKey(),
            app=APP_NAME,
        )

    def oauth_call_url(self):
        host = (
            self.callerArgs.data.get("endpoint_token_oauth_credentials", [None])[0]
            or self.callerArgs.data.get("endpoint_token", [None])[0]
            or self.callerArgs.data.get("endpoint", [None])[0]
        )

        return f"https://{host}/{TOKEN_ENDPOINT.lstrip('/')}"

    def oauth_client_credentials_call(self):
        auth_type = ${auth_type}
        if auth_type != "oauth_client_credentials":
            return

        client_id = (
            self.callerArgs.data.get("client_id_oauth_credentials", [None])[0]
            or self.callerArgs.data.get("client_id", [None])[0]
        )

        client_secret = (
            self.callerArgs.data.get("client_secret_oauth_credentials", [None])[0]
            or self.callerArgs.data.get("client_secret", [None])[0]
        )

        params = {
            "grant_type": "client_credentials",
            "client_id": client_id,
            "client_secret": client_secret,
            "url": self.oauth_call_url(),
            "method": "POST",
        }

        if "scope" in self.callerArgs.data:
            params["scope"] = self.callerArgs.data.get("scope", [None])[0]

        data = json.loads(
            self._rest_client.post(
                self._oauth_url,
                body=params,
                headers=[("Content-Type", "application/json")],
                output_mode="json",
            ).body.read().decode("utf-8")
        )["entry"][0]["content"]

        if "access_token" not in data:
            data = data.get("error", data)
            raise InternalException("Error while trying to obtain OAuth token: %s" % data)

        self.payload["access_token"] = data["access_token"]

        for key in ["refresh_token", "instance_url"]:
            if key in data:
                self.payload[key] = data[key]

    def handleCreate(self, confInfo):
        self.oauth_client_credentials_call()
        return super().handleCreate(confInfo)

    def handleEdit(self, confInfo):
        self.oauth_client_credentials_call()
        return super().handleEdit(confInfo)
"""

    def __init__(
        self,
        name: Optional[str],
        namespace: str,
        app_name: str,
        token_endpoint: str,
        auth_condition: bool,
        **kwargs: Any,
    ):
        super().__init__(name, namespace, **kwargs)
        self.token_endpoint = token_endpoint
        self.app_name = app_name
        self.auth_condition = auth_condition

    def generate_rh(self) -> str:
        entity = self._entities[0]

        oauth_endpoint = f"{self._namespace}_oauth"
        class_name = "HandlerWithOauth"

        if self.auth_condition:
            auth_type = 'self.callerArgs.data.get("auth_type", [""])[0]'
        else:
            auth_type = '"oauth_client_credentials"'

        cls_content = Template(self._cls_template).substitute(
            base_class=self.rh_class,
            oauth_endpoint=repr(oauth_endpoint),
            class_name=class_name,
            token_endpoint=repr(self.token_endpoint),
            app_name=repr(self.app_name),
            auth_type=auth_type,
        )

        return self._rh_template.format(
            handler_module=self.rh_module,
            handler_class=self.rh_class,
            handler_class_used=class_name,
            entity=entity.generate_rh(),
            conf_name=self.conf_name,
            config_name=self._name,
            need_reload=self.need_reload,
            additional_code=cls_content,
            additional_imports=self._imports,
        )



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional, Any

from splunk_add_on_ucc_framework.entity.entity import Entity


class IndexEntity(Entity):
    def long_form(self) -> dict[str, Any]:
        definition = {
            "type": "singleSelect",
            "field": self["field"],
            "label": self["label"],
            "defaultValue": "default",
            "options": {
                "endpointUrl": "data/indexes?search=isInternal=0+disabled=0",
                "denyList": "^_.*$",
                "createSearchChoice": True,
            },
            "validators": [
                {
                    "type": "regex",
                    "errorMsg": "Index names must begin with a letter or a number "
                    "and must contain only letters, numbers, underscores or hyphens.",
                    "pattern": r"^[a-zA-Z0-9][a-zA-Z0-9\\_\\-]*$",
                },
                {
                    "type": "string",
                    "errorMsg": "Length of index name should be between 1 and 80.",
                    "minLength": 1,
                    "maxLength": 80,
                },
            ],
        }

        for name in ("help", "required", "defaultValue"):
            if name in self:
                definition[name] = self[name]

        return definition

    @classmethod
    def from_definition(cls, definition: dict[str, Any]) -> Optional[Entity]:
        if definition.get("type") == "index":
            return cls(definition)

        return None



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import shutil
from os import path
from typing import Any, Optional

from splunk_add_on_ucc_framework.commands.modular_alert_builder import normalize
from splunk_add_on_ucc_framework.generators.file_generator import FileGenerator
from splunk_add_on_ucc_framework.global_config import GlobalConfig


class AlertActionsConf(FileGenerator):
    __description__ = (
        "Generates `alert_actions.conf` and `alert_actions.conf.spec` file "
        "for the custom alert actions defined in globalConfig"
    )

    def __init__(
        self, global_config: GlobalConfig, input_dir: str, output_dir: str
    ) -> None:
        super().__init__(global_config, input_dir, output_dir)
        self.conf_file = "alert_actions.conf"
        self.conf_spec_file = f"{self.conf_file}.spec"

        envs = normalize.normalize(
            global_config.alerts,
            global_config.namespace,
        )
        schema_content = envs["schema.content"]
        self._alert_settings = schema_content["modular_alerts"]

        deny_list = frozenset(
            [
                "short_name",
                "alert_props",
                "parameters",
                "uuid",
                "code",
                "largeIcon",
                "smallIcon",
                "index",
                "iconFileName",  # it is a config from globalConfig that gets written to icon_path
                "customScript",  # it is a config from globalConfig only for Python script
            ]
        )
        _router = {
            "dropdownlist": "list",
            "text": "string",
            "textarea": "string",
            "checkbox": "bool",
            "password": "password",
            "dropdownlist_splunk_search": "list",
            "radio": "list",
        }

        self.alerts: dict[str, Any] = {}
        self.alerts_spec: dict[str, Any] = {}

        for alert in self._alert_settings:
            alert_name = alert["short_name"]
            self.alerts[alert_name] = []
            self.alerts_spec[alert_name] = []
            # process the 'iconFileName' property for alert actions
            if alert.get("iconFileName", "alerticon.png") != "alerticon.png":
                self.alerts[alert_name].append(f"icon_path = {alert['iconFileName']}")
            else:
                self.alerts[alert_name].append("icon_path = alerticon.png")
                # we copy UCC framework's alerticon.png only when a custom isn't provided
                shutil.copy(
                    path.join(self._ucc_dir, "static", "alerticon.png"),
                    path.join(self._get_output_dir(), "appserver", "static"),
                )
            # process alert action properties in bulk
            for k, v in alert.items():
                if k == "adaptive_response":
                    new_cam = {
                        sub_k: sub_v
                        for sub_k, sub_v in list(v.items())
                        if sub_k != "sourcetype" and sub_v
                    }
                    value = f"param._cam = {json.dumps(new_cam)}"
                    self.alerts[alert_name].append(value)
                    self.alerts_spec[alert_name].append(
                        "param._cam = <json> Adaptive Response parameters."
                    )
                elif k == "parameters":
                    for param in v:
                        param_name = param["name"].strip()
                        if param.get("default_value") is not None:
                            self.alerts[alert_name].append(
                                f"param.{param_name} = {str(param['default_value']).strip()}"
                            )
                        else:
                            self.alerts[alert_name].append(f"param.{param_name} = ")

                        # fetching details for alert_actions.conf.spec file
                        format_type = _router[param["format_type"]]
                        is_required = (
                            "It's a required parameter."
                            if param.get("required") and param["required"]
                            else ""
                        )
                        param_default_value = param.get("default_value")
                        default_value = (
                            f"It's default value is {param_default_value}."
                            if param_default_value
                            else ""
                        )
                        value = (
                            f'param.{param["name"]} = <{format_type}> '
                            f'{param["label"]}. {is_required} {default_value}'
                        )
                        self.alerts_spec[alert_name].append(value)
                elif k not in deny_list:
                    value = f"{str(k).strip()} = {str(v).strip()}"
                    self.alerts[alert_name].append(value)

    def generate(self) -> Optional[list[dict[str, str]]]:
        conf_files: list[dict[str, str]] = []
        conf = self.generate_conf()
        conf_spec = self.generate_conf_spec()
        if conf is not None:
            conf_files.append(conf)
        if conf_spec is not None:
            conf_files.append(conf_spec)
        return None if conf_files == [] else conf_files

    def generate_conf(self) -> Optional[dict[str, str]]:
        if not self.alerts:
            return None

        file_path = self.get_file_output_path(["default", self.conf_file])
        self.set_template_and_render(
            template_file_path=["conf_files"], file_name="alert_actions_conf.template"
        )
        rendered_content = self._template.render(alerts=self.alerts)
        return {
            "file_name": self.conf_file,
            "file_path": file_path,
            "content": rendered_content,
        }

    def generate_conf_spec(self) -> Optional[dict[str, str]]:
        if not self.alerts_spec:
            return None

        file_path = self.get_file_output_path(["README", self.conf_spec_file])
        self.set_template_and_render(
            template_file_path=["README"],
            file_name="alert_actions_conf_spec.template",
        )
        rendered_content = self._template.render(alerts=self.alerts_spec)
        return {
            "file_name": self.conf_spec_file,
            "file_path": file_path,
            "content": rendered_content,
        }



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import defaultdict
from typing import Any, Optional

from splunk_add_on_ucc_framework.generators.file_generator import FileGenerator
from splunk_add_on_ucc_framework.global_config import GlobalConfig


class InputsConf(FileGenerator):
    __description__ = (
        "Generates `inputs.conf` and `inputs.conf.spec` "
        "file for the services mentioned in globalConfig"
    )

    def __init__(
        self, global_config: GlobalConfig, input_dir: str, output_dir: str
    ) -> None:
        super().__init__(global_config, input_dir, output_dir)
        self.conf_file = self._conf_file_name("inputs")

        # A list of service names from globalConfig that will be in inputs.conf
        self.inputs_conf_names: list[str] = []
        # A dictionary of dictionaries of default properties for each service in inputs.conf
        self.inputs_conf_params: dict[str, dict[str, Any]] = defaultdict(
            lambda: defaultdict(dict)
        )

        # A dictionary of lists of properties for each service in inputs.conf
        self.inputs_conf_spec: dict[str, list[str]] = defaultdict(list)
        # A dictionary of lists of properties for each service in their own spec files
        # (i.e. dict key is the spec file name)
        self.other_spec_files: dict[str, list[str]] = defaultdict(list)

        if not global_config.has_inputs():
            return

        for service in global_config.inputs:
            default_values = None

            # If the service has a conf property, it will be in a separate spec file
            # Otherwise, it will be in inputs.conf
            if service.get("conf") is not None:
                spec_properties = self.other_spec_files[service["conf"]]
            else:
                spec_properties = self.inputs_conf_spec[service["name"]]
                default_values = self.inputs_conf_params[service["name"]]

            # Always add the service name to the list of service names in inputs.conf
            # to add at least the Python version
            self.inputs_conf_names.append(service["name"])

            if default_values is not None and service.get("disableNewInput"):
                default_values["disabled"] = "true"

            for entity in service.get("entity", []):
                field_name = entity["field"]

                # Skip the name field as it is already in the service name
                if field_name == "name":
                    continue

                # Construct the property spec description
                field_value_parts = []

                if entity.get("help"):
                    field_value_parts.append(entity["help"].replace("\n", " "))

                field_default_value = entity.get("defaultValue")

                if field_default_value:
                    # Convert boolean values to lowercase strings
                    if type(field_default_value) is bool:
                        value = str(field_default_value).lower()
                    else:
                        value = str(field_default_value)

                    # Add the default value to the description and dictionary
                    if default_values is not None:
                        default_values[field_name] = value
                    field_value_parts.append(f"(Default: {value})")

                field_value = " ".join(field_value_parts)
                prop = f"{field_name} = {field_value}".rstrip()
                spec_properties.append(prop)

    def _conf_file_name(self, conf_name: str) -> str:
        return f"{conf_name}.conf"

    def _spec_file_name(self, conf_name: str) -> str:
        return f"{self._conf_file_name(conf_name)}.spec"

    def generate(self) -> Optional[list[dict[str, str]]]:
        conf_files: list[dict[str, str]] = []
        conf = self.generate_conf()
        conf_spec = self.generate_conf_spec()
        if conf is not None:
            conf_files.append(conf)
        if conf_spec is not None:
            conf_files.extend(conf_spec)
        return None if conf_files == [] else conf_files

    def generate_conf(self) -> Optional[dict[str, str]]:
        if not self.inputs_conf_names:
            return None

        file_path = self.get_file_output_path(["default", self.conf_file])
        self.set_template_and_render(
            template_file_path=["conf_files"], file_name="inputs_conf.template"
        )

        rendered_content = self._template.render(
            input_names=self.inputs_conf_names,
            default_values=self.inputs_conf_params,
        )
        return {
            "file_name": self.conf_file,
            "file_path": file_path,
            "content": rendered_content,
        }

    def _generate_spec_inputs(self) -> Optional[dict[str, str]]:
        if not self.inputs_conf_spec:
            return None

        spec_file = self._spec_file_name("inputs")
        file_path = self.get_file_output_path(["README", spec_file])

        self.set_template_and_render(
            template_file_path=["README"], file_name="inputs_conf_spec.template"
        )

        rendered_content = self._template.render(
            input_names=self.inputs_conf_names,
            input_stanzas=self.inputs_conf_spec,
        )
        return {
            "file_name": spec_file,
            "file_path": file_path,
            "content": rendered_content,
        }

    def _generate_spec_other(self, name: str, parameters: list[str]) -> dict[str, str]:
        spec_file = self._spec_file_name(name)
        file_path = self.get_file_output_path(["README", spec_file])

        content = ["[<name>]"]
        content.extend(parameters)

        return {
            "file_name": spec_file,
            "file_path": file_path,
            "content": "\n".join(content),
        }

    def generate_conf_spec(self) -> Optional[list[dict[str, str]]]:
        files: list[dict[str, str]] = []
        spec_input = self._generate_spec_inputs()
        if spec_input is not None:
            files.append(spec_input)

        for name, params in self.other_spec_files.items():
            files.append(self._generate_spec_other(name, params))

        if not files:
            return None
        return files



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
from abc import ABC
from os.path import realpath, sep, dirname, abspath
from typing import Any, Union, Optional

from jinja2 import Environment, FileSystemLoader, select_autoescape

from splunk_add_on_ucc_framework.utils import (
    write_file,
)
from splunk_add_on_ucc_framework.commands.rest_builder.global_config_builder_schema import (
    GlobalConfigBuilderSchema,
)
from splunk_add_on_ucc_framework.global_config import GlobalConfig
from splunk_add_on_ucc_framework import __file__ as ucc_framework_file
from . import file_const as fc

__all__ = ["FileGenerator", "begin"]

logger = logging.getLogger("ucc_gen")


class FileGenerator(ABC):
    __description__ = "DESCRIBE THE FILE THAT IS GENERATED"

    _ucc_dir = abspath(dirname(ucc_framework_file))

    def __init__(
        self,
        global_config: GlobalConfig,
        input_dir: str,
        output_dir: str,
    ) -> None:
        """
        :param global_config: the GlobalConfig object that is validated and parsed
        :param input_dir: the path to the source code of globalConfig.(json|yaml)
        :param output_dir: the path to output/<addon_name> directory

        """
        super().__init__()
        self._global_config = global_config
        self._input_dir = input_dir
        self._output_dir = output_dir
        self._template_dir = [(sep.join([self._ucc_dir, "templates"]))]
        self._template = Environment(
            loader=FileSystemLoader(self._template_dir),
            trim_blocks=True,
            lstrip_blocks=True,
            keep_trailing_newline=True,
        )
        self._addon_name: str = global_config.product
        self.writer = write_file
        self._gc_schema: GlobalConfigBuilderSchema = GlobalConfigBuilderSchema(
            global_config
        )

    def generate(self) -> Optional[list[dict[str, str]]]:
        raise NotImplementedError()

    def _get_output_dir(self) -> str:
        return sep.join([realpath(self._output_dir), self._addon_name])

    def get_file_output_path(self, output_piece: Union[list[str], str]) -> str:
        if isinstance(output_piece, str):
            return sep.join([self._get_output_dir(), output_piece])
        elif isinstance(output_piece, list):
            return sep.join([self._get_output_dir()] + output_piece)

        raise TypeError(
            "Invalid type of output_piece, provided type='%s'" % (type(output_piece))
        )

    def set_template_and_render(
        self, template_file_path: list[str], file_name: str
    ) -> None:
        assert file_name.endswith(".template")
        select_autoescape(disabled_extensions=("template"))

        self._template = Environment(
            loader=FileSystemLoader(sep.join(self._template_dir + template_file_path)),
            trim_blocks=True,
            lstrip_blocks=True,
            keep_trailing_newline=True,
        )
        self._template = self._template.get_template(file_name)

    def _render(self, file_name: str, **kwargs: Any) -> str:
        select_autoescape(disabled_extensions=("template",))
        template = self._template.get_template(file_name)
        return template.render(kwargs)


def begin(
    global_config: GlobalConfig, input_dir: str, output_dir: str
) -> list[dict[str, str]]:
    generated_files: list[dict[str, str]] = []
    for item in fc.GEN_FILE_LIST:
        file_details = item.file_class(global_config, input_dir, output_dir).generate()
        if file_details is None:
            continue
        for details in file_details:
            write_file(
                details["file_name"],
                details["file_path"],
                details["content"],
                merge_mode=details.get("merge_mode", "stanza_overwrite"),
            )
            logger.info(
                f"Successfully generated '{details['file_name']}' at '{details['file_path']}"
            )
            generated_files.append({details["file_name"]: details["file_path"]})

    return generated_files



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import re
from os import path as op
from typing import Any

from jinja2 import Environment, FileSystemLoader, select_autoescape

from splunk_add_on_ucc_framework.commands.modular_alert_builder import (
    arf_consts as ac,
)
from splunk_add_on_ucc_framework.utils import write_file

logger = logging.getLogger("ucc_gen")


class AlertActionsPyGenerator:
    def __init__(
        self,
        addon_name: str,
        input_setting: dict[str, Any],
        package_path: str,
    ) -> None:
        self._addon_name = addon_name
        self._all_setting = input_setting
        self._package_path = package_path
        self._alert_actions_setting = input_setting[ac.MODULAR_ALERTS]
        ta_name = input_setting.get(ac.SHORT_NAME)
        if ta_name is None:
            raise ValueError(
                f"{ac.SHORT_NAME} key should be present in input_settings passed"
            )
        space_replace = re.compile(r"[^\w]+")
        self._lib_dir = space_replace.sub("_", ta_name.lower())
        self._templates = Environment(
            loader=FileSystemLoader(
                op.join(op.dirname(op.realpath(__file__)), "arf_template")
            ),
            trim_blocks=True,
            lstrip_blocks=True,
            keep_trailing_newline=True,
            autoescape=select_autoescape(disabled_extensions=("template")),
        )

    def _get_alert_py_name(self, alert: Any) -> str:
        return alert[ac.SHORT_NAME] + ".py"

    def _get_alert_py_path(self, alert: Any) -> str:
        return op.join(self._package_path, "bin", self._get_alert_py_name(alert))

    def _get_alert_helper_py_name(self, alert: Any) -> str:
        return "modalert_" + alert[ac.SHORT_NAME] + "_helper.py"

    def _get_alert_helper_py_path(self, alert: Any) -> str:
        return op.join(
            self._package_path,
            "bin",
            self._lib_dir,
            self._get_alert_helper_py_name(alert),
        )

    def gen_main_py_file(self, alert: Any) -> None:
        template = self._templates.get_template("alert_action.py.template")
        rendered_content = template.render(
            addon_name=self._addon_name,
            lib_name=self._lib_dir,
            mod_alert=alert,
            helper_name=op.splitext(self._get_alert_helper_py_name(alert))[0],
            custom_logic_handler=alert.get("customScript", ""),
        )
        logger.debug(f"Writing to file {self._get_alert_py_path(alert)}")
        write_file(
            self._get_alert_py_name(alert),
            self._get_alert_py_path(alert),
            rendered_content,
        )

    def gen_helper_py_file(self, alert: Any) -> None:
        template = self._templates.get_template("alert_action_helper.py.template")
        rendered_content = template.render(
            input=self._all_setting,
            mod_alert=alert,
        )
        logger.debug(f"Writing to file {self._get_alert_py_path(alert)}")
        if not alert.get("customScript"):
            # we only write the template file if the custom Script isn't available.
            write_file(
                self._get_alert_helper_py_name(alert),
                self._get_alert_helper_py_path(alert),
                rendered_content,
            )

    def handle(self) -> None:
        for alert in self._alert_actions_setting:
            logger.info(
                f"Generating Python file for alert action {alert[ac.SHORT_NAME]}"
            )
            self.gen_main_py_file(alert)
            self.gen_helper_py_file(alert)



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import csv
import glob
import logging
import os
import re
import stat
import subprocess
import sys
from pathlib import Path

from packaging.requirements import Requirement, InvalidRequirement
from packaging.version import Version
from typing import Optional
from collections.abc import Iterable
from splunk_add_on_ucc_framework.global_config import OSDependentLibraryConfig

logger = logging.getLogger("ucc_gen")


LIBS_REQUIRED_FOR_UI = {"splunktaucclib": "6.6.0"}
LIBS_REQUIRED_FOR_OAUTH = {"solnlib": "5.5.0"}


class SplunktaucclibNotFound(Exception):
    pass


class WrongSplunktaucclibVersion(Exception):
    pass


class WrongSolnlibVersion(Exception):
    pass


class CouldNotInstallRequirements(Exception):
    pass


class InvalidArguments(Exception):
    pass


def _subprocess_run(
    command: str,
    command_desc: Optional[str] = None,
    env: Optional[dict[str, str]] = None,
) -> "subprocess.CompletedProcess[bytes]":
    command_desc = command_desc or command
    try:
        logger.info(f"Executing: {command}")
        process_result = subprocess.run(
            command, shell=True, env=env, capture_output=True
        )
        return_code = process_result.returncode
        if return_code < 0:
            logger.error(
                f"Child ({command_desc}) was terminated by signal {-return_code}"
            )
        if return_code > 0:
            logger.error(f"Command ({command_desc}) returned {return_code} status code")
        return process_result
    except OSError as e:
        logger.error(f"Execution ({command_desc}) failed due to {e}")
        raise e


def _pip_install(installer: str, command: str, command_desc: str) -> None:
    cmd = f"{installer} -m pip install {command}"
    try:
        subprocess_result = _subprocess_run(command=cmd, command_desc=command_desc)
        return_code = subprocess_result.returncode
        if return_code != 0:
            raise CouldNotInstallRequirements(subprocess_result.stderr.decode())
    except OSError as e:
        raise CouldNotInstallRequirements(e) from e


def _pip_is_lib_installed(
    installer: str,
    target: str,
    libname: str,
    version: Optional[str] = None,
    allow_higher_version: bool = False,
) -> bool:
    if not version and allow_higher_version:
        raise InvalidArguments(
            "Parameter 'allow_higher_version' can not be set to True if 'version' parameter is not provided"
        )

    lib_installed_cmd = f"{installer} -m pip show --version {libname}"

    try:
        my_env = os.environ.copy()
        my_env["PYTHONPATH"] = target

        # Disable writing of .pyc files (__pycache__)
        my_env["PYTHONDONTWRITEBYTECODE"] = "1"

        result = _subprocess_run(command=lib_installed_cmd, env=my_env)
        if result.returncode != 0 or "Version:" not in result.stdout.decode("utf-8"):
            logger.error(
                f"Command result: {result.stdout.decode()} {result.stderr.decode()}"
            )
            return False

        if version:
            pip_show_result = result.stdout.decode("utf-8").splitlines()
            result_row = next(el for el in pip_show_result if el.startswith("Version:"))
            result_version = result_row.split("Version:")[1].strip()
            if allow_higher_version:
                logger.info(
                    f"Command result: {result.stdout.decode()} {result.stderr.decode()}"
                )
                return Version(result_version) >= Version(version)
            return Version(result_version) == Version(version)
        else:
            return result.returncode == 0

    except OSError as exc:
        logger.error(f"Command execution failed with error message: {str(exc)}")
        raise CouldNotInstallRequirements from exc


def _check_libraries_required_for_oauth(
    python_binary_name: str, ucc_lib_target: str, path_to_requirements_file: str
) -> None:
    for lib, version in LIBS_REQUIRED_FOR_OAUTH.items():
        if not _pip_is_lib_installed(
            installer=python_binary_name,
            target=ucc_lib_target,
            libname=lib,
            version=version,
            allow_higher_version=True,
        ):
            raise WrongSolnlibVersion(
                f"{lib} found at {path_to_requirements_file}, but is not of latest version."
                f" Please make sure {lib} is of version greater than or equal to {version}."
            )


def _check_libraries_required_for_ui(
    python_binary_name: str, ucc_lib_target: str, path_to_requirements_file: str
) -> None:
    for lib, version in LIBS_REQUIRED_FOR_UI.items():
        if not _pip_is_lib_installed(
            installer=python_binary_name,
            target=ucc_lib_target,
            libname=lib,
        ):
            raise SplunktaucclibNotFound(
                f"This add-on has an UI, so the {lib} is required but not found in "
                f"{path_to_requirements_file}. Please add it there and make sure it is at least version {version}."
            )
        if not _pip_is_lib_installed(
            installer=python_binary_name,
            target=ucc_lib_target,
            libname=lib,
            version=version,
            allow_higher_version=True,
        ):
            raise WrongSplunktaucclibVersion(
                f"{lib} found but has the wrong version. Please make sure it is at least version {version}."
            )


def parse_excludes(excludes_path: str) -> Optional[list[str]]:
    if not os.path.isfile(excludes_path):
        return None

    with open(excludes_path) as fp:
        excludes = fp.read()

    # https://pip.pypa.io/en/latest/reference/requirements-file-format/#comments
    comment_start = re.compile(r"\s#")

    excluded_libs = []

    for exclude_str in excludes.splitlines():
        exclude_str = comment_start.split(exclude_str, maxsplit=1)[0].strip()

        if not exclude_str or exclude_str.startswith("#"):
            continue

        try:
            exclude = Requirement(exclude_str)
        except InvalidRequirement:
            raise InvalidArguments(
                f"Exclusion '{exclude_str}' is not a valid requirement."
            )

        if exclude.specifier or exclude.marker or exclude.url or exclude.extras:
            raise InvalidArguments(
                f"Exclusion '{exclude_str}' should not contain a version specifier, marker, URL, or extras."
            )

        excluded_libs.append(str(exclude))

    if not excluded_libs:
        return None

    return excluded_libs


def install_python_libraries(
    source_path: str,
    ucc_lib_target: str,
    python_binary_name: str,
    includes_ui: bool = False,
    os_libraries: Optional[list[OSDependentLibraryConfig]] = None,
    pip_version: str = "latest",
    pip_legacy_resolver: bool = False,
    pip_custom_flag: Optional[str] = None,
    includes_oauth: bool = False,
) -> None:
    path_to_requirements_file = os.path.join(source_path, "lib", "requirements.txt")
    path_to_excludes = os.path.join(source_path, "lib", "exclude.txt")
    excludes = parse_excludes(path_to_excludes)

    if os.path.isfile(path_to_requirements_file):
        logger.info(f"Installing requirements from {path_to_requirements_file}")
        if not os.path.exists(ucc_lib_target):
            os.makedirs(ucc_lib_target)
        install_libraries(
            requirements_file_path=path_to_requirements_file,
            installation_path=ucc_lib_target,
            installer=python_binary_name,
            pip_version=pip_version,
            pip_legacy_resolver=pip_legacy_resolver,
            pip_custom_flag=pip_custom_flag,
        )
        if excludes:
            logger.info(f"Excluding libraries: {', '.join(excludes)}")
            remove_packages(installation_path=ucc_lib_target, package_names=excludes)
        if includes_ui:
            _check_libraries_required_for_ui(
                python_binary_name, ucc_lib_target, path_to_requirements_file
            )
        if includes_oauth:
            _check_libraries_required_for_oauth(
                python_binary_name, ucc_lib_target, path_to_requirements_file
            )

        cleanup_libraries = install_os_dependent_libraries(
            ucc_lib_target=ucc_lib_target,
            installer=python_binary_name,
            os_libraries=os_libraries,
        )

        packages_to_remove = {
            "setuptools",
            "bin",
            "pip",
            "distribute",
            "wheel",
        }
        # we can remove os-dependent libraries from the installation_path to save some space.
        packages_to_remove.update(cleanup_libraries)
        remove_packages(
            installation_path=ucc_lib_target,
            package_names=packages_to_remove,
        )

        remove_execute_bit(ucc_lib_target)
    else:
        logger.warning(
            f"Could not find requirements file @ {path_to_requirements_file}, nothing to install"
        )


def install_libraries(
    requirements_file_path: str,
    installation_path: str,
    installer: str,
    pip_version: str = "latest",
    pip_legacy_resolver: bool = False,
    pip_custom_flag: Optional[str] = None,
) -> None:
    """
    Upgrades `pip` version to the latest one and installs requirements to the
    specified path.
    """

    if pip_version == "latest":
        pip_update_command = "--upgrade pip"
    else:
        pip_update_command = f"--upgrade pip=={pip_version.strip()}"

    if pip_version.strip() == "23.2" and pip_legacy_resolver:
        logger.error(
            "You cannot use the legacy resolver with pip 23.2. "
            "Please remove '--pip-legacy-resolver' from your build command or "
            "use a different version of pip e.g. 23.2.1"
        )
        sys.exit(1)

    deps_resolver = "--use-deprecated=legacy-resolver " if pip_legacy_resolver else ""
    custom_flag = (
        pip_custom_flag
        if pip_custom_flag
        else "--no-compile --prefer-binary --ignore-installed "
    )
    pip_install_command = (
        f'-r "{requirements_file_path}" '
        f"{deps_resolver}"
        f'--target "{installation_path}" '
        f"{custom_flag}"
    )
    try:
        _pip_install(
            installer=installer, command=pip_update_command, command_desc="pip upgrade"
        )

        _pip_install(
            installer=installer, command=pip_install_command, command_desc="pip install"
        )
    except CouldNotInstallRequirements as exc:
        logger.error(f"Command execution failed with error message: {str(exc)}")
        sys.exit(1)


def determine_record_separator(paths: list[str], dist_info: str) -> str:
    """
    Determines the separator used in RECORD files for the given dist-info directory.
    As Windows can use both '/' and '\' as path separators, we need to check the RECORD files
    to determine which one is used.
    If no RECORD files are found, defaults to '/'.
    """
    dist_info_dir_escaped = re.escape(dist_info)
    record_pattern = re.compile(rf"^{dist_info_dir_escaped}([\\/])RECORD$")

    for path in paths:
        separator_match = record_pattern.match(path)

        if separator_match:
            return separator_match.group(1)

    return "/"


def _remove_package(installation_path: str, package_name: str) -> bool:
    name_sub = re.compile(r"[^\w\d.]+", re.UNICODE)

    # https://peps.python.org/pep-0491/
    def sub(name: str) -> str:
        return name_sub.sub("_", name)

    # RECORD files are located in the *.dist-info directories
    # Example content:
    #     requests-2.32.5.dist-info/INSTALLER,sha256=zuuue4knoyJ-UwPPXg8fezS7VCrXJQrAP7zeNuwvFQg,4
    #     requests-2.32.5.dist-info/METADATA,sha256=ZbWgjagfSRVRPnYJZf8Ut1GPZbe7Pv4NqzZLvMTUDLA,4945
    #     requests-2.32.5.dist-info/RECORD,,
    #     requests/__init__.py,sha256=4xaAERmPDIBPsa2PsjpU9r06yooK-2mZKHTZAhWRWts,5072

    exclude_safe = sub(package_name)
    record_files = os.path.join(
        installation_path, f"{exclude_safe}-*.dist-info", "RECORD"
    )

    is_deleted = False

    for record in glob.glob(record_files):
        top_level_dirs = set()

        dist_info_dir = os.path.basename(os.path.dirname(record))

        with open(record) as fp:
            paths = [line[0] for line in csv.reader(fp)]

        separator = determine_record_separator(paths, dist_info_dir)

        for path in paths:
            if os.path.isabs(path):
                raise ValueError(
                    f"Absolute paths are not allowed in RECORD files: {path}"
                )
            os.remove(os.path.join(installation_path, path))
            top_level_dirs.add(path.split(separator)[0])

        for directory in top_level_dirs:
            for path, _, _ in os.walk(
                os.path.join(installation_path, directory), topdown=False
            ):
                if not os.listdir(path):
                    os.rmdir(path)

        deleted_folders = set()
        not_deleted_folders = set()

        for folder in top_level_dirs:
            if os.path.isdir(os.path.join(installation_path, folder)):
                not_deleted_folders.add(folder)
            else:
                deleted_folders.add(folder)

        deleted_folders_str = ", ".join(
            f"{folder} (deleted)" for folder in deleted_folders
        )
        not_deleted_folders_str = ", ".join(
            f"{folder} (not deleted)" for folder in not_deleted_folders
        )

        logger.info(
            "Excluded package %s with directories: %s",
            package_name,
            ", ".join((deleted_folders_str, not_deleted_folders_str)),
        )
        is_deleted = True

    return is_deleted


def remove_packages(installation_path: str, package_names: Iterable[str]) -> None:
    package_names = set(package_names)

    logger.info(f"Removing {len(package_names)} package(s): {', '.join(package_names)}")

    successful_removes = set()

    for package_name in package_names:
        if _remove_package(installation_path, package_name):
            successful_removes.add(package_name)

    remaining_packages = package_names - successful_removes

    if remaining_packages:
        logger.info(
            f"Could not find directories for {len(remaining_packages)} packages(s) to delete: "
            f"{', '.join(remaining_packages)}"
        )


def remove_execute_bit(installation_path: str) -> None:
    p = Path(installation_path)
    no_user_exec = ~stat.S_IEXEC
    no_group_exec = ~stat.S_IXGRP
    no_other_exec = ~stat.S_IXOTH
    no_exec = no_user_exec & no_group_exec & no_other_exec

    for o in p.rglob("*"):
        if not o.is_dir() and os.access(o, os.X_OK):
            logger.info(f"  fixing {o} execute bit")
            current_permissions = stat.S_IMODE(os.lstat(o).st_mode)
            os.chmod(o, current_permissions & no_exec)


def install_os_dependent_libraries(
    ucc_lib_target: str,
    installer: str,
    os_libraries: Optional[list[OSDependentLibraryConfig]],
) -> set[str]:
    cleanup_libraries: set[str] = set()

    if not os_libraries:
        logger.info("No os-dependentLibraries to install.")
        return cleanup_libraries

    logger.info("Installing os-dependentLibraries.")

    validate_conflicting_paths(os_libraries)
    for os_lib in os_libraries:
        if os_lib.dependencies is False and not _pip_is_lib_installed(
            installer=installer,
            target=ucc_lib_target,
            libname=os_lib.name,
            version=os_lib.version,
        ):
            logger.error(
                f"""
OS dependent library {os_lib.name} = {os_lib.version} SHOULD be defined in requirements.txt.
When the os dependent library is installed without its dependencies it has to be listed in requirements.txt.
Possible solutions, either:
1. os-dependentLibraries.name[{os_lib.name}].dependencies = True
2. Add {os_lib.name}=={os_lib.version} in requirements.txt
"""
            )
            raise CouldNotInstallRequirements

        target_path = os.path.join(ucc_lib_target, os.path.normpath(os_lib.target))
        if not os.path.exists(target_path):
            os.makedirs(target_path)

        pip_download_command = (
            f"{os_lib.deps_flag} "
            f"--no-compile "
            f"--platform {os_lib.platform} "
            f"--python-version {os_lib.python_version} "
            f"--target {target_path}"
            f" --only-binary=:all: "
            f"{os_lib.name}=={os_lib.version} "
            f"{os_lib.ignore_requires_python} "
        )

        try:
            _pip_install(
                installer=installer,
                command=pip_download_command,
                command_desc="pip download",
            )
        except CouldNotInstallRequirements as exc:
            logger.error(
                "Downloading process failed. Please verify parameters in the globalConfig.json file."
            )
            sys.exit(f"Package build aborted with error message: {str(exc)}")
        cleanup_libraries.add(os_lib.name)
    return cleanup_libraries


def validate_conflicting_paths(libs: list[OSDependentLibraryConfig]) -> bool:
    name_target_pairs = [(lib.name, lib.target) for lib in libs]
    conflicts = {x for x in name_target_pairs if name_target_pairs.count(x) > 1}
    if conflicts:
        logger.error(
            f"Found conflicting paths for libraries: {conflicts}. "
            "Please make sure that the paths are unique."
        )
        raise CouldNotInstallRequirements
    return True



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import argparse
import re
import sys
from typing import Optional
from collections.abc import Sequence
import logging

from splunk_add_on_ucc_framework.commands import build
from splunk_add_on_ucc_framework.commands import init
from splunk_add_on_ucc_framework.commands import import_from_aob
from splunk_add_on_ucc_framework.commands import package
from splunk_add_on_ucc_framework.commands import validate
from splunk_add_on_ucc_framework.commands import publish

logger = logging.getLogger("ucc_gen")


# This is a necessary change to have, so we don't release a breaking change.
# This class adds a default subparser if none specified. In our case it will be
# `build`subclass.
# This is a temporary change and everyone migrates to `ucc-gen build` instead of
# just using `ucc-gen` this will be removed.
# The limitation of this approach is that we can't have global options without a
# subparser being specified. Example is `--version`, the default subparser will
# be added here as well. But this is not a big deal for now, we don't have
# global options anyway.


class DefaultSubcommandArgumentParser(argparse.ArgumentParser):
    __default_subparser = None

    def set_default_subparser(self, name: str) -> None:
        self.__default_subparser = name

    def _parse_known_args(self, arg_strings, *args, **kwargs):  # type: ignore
        in_args = set(arg_strings)
        d_sp = self.__default_subparser
        if d_sp is not None and not {"-h", "--help"}.intersection(in_args):
            for x in self._subparsers._actions:  # type: ignore
                subparser_found = (
                    isinstance(x, argparse._SubParsersAction)
                    and len(arg_strings) > 0
                    and arg_strings[0] in x._name_parser_map.keys()
                )
                if subparser_found:
                    break
            else:
                logger.warning(
                    "Please use `ucc-gen build` if you want to build "
                    "an add-on, using just `ucc-gen` will be deprecated"
                )
                arg_strings = [d_sp] + arg_strings
        return super()._parse_known_args(arg_strings, *args, **kwargs)


def main(argv: Optional[Sequence[str]] = None) -> int:
    argv = argv if argv is not None else sys.argv[1:]
    parser = DefaultSubcommandArgumentParser(prog="ucc-gen")
    parser.set_default_subparser("build")
    subparsers = parser.add_subparsers(dest="command")

    build_parser = subparsers.add_parser("build", description="Build an add-on")
    build_parser.add_argument(
        "--source",
        type=str,
        nargs="?",
        help="folder containing the app.manifest and app source",
        default="package",
    )
    build_parser.add_argument(
        "--config",
        type=config_file_type,
        nargs="?",
        help="path to configuration file, defaults to globalConfig file in parent directory of source provided. "
        "Only .json and .yaml files are accepted.",
        default=None,
    )
    build_parser.add_argument(
        "--ta-version",
        type=str,
        help="version of add-on, default version is version specified in the "
        "package such as app.manifest, app.conf, and globalConfig file",
        default=None,
    )
    build_parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="output path to store built add-on",
        default=None,
    )
    build_parser.add_argument(
        "--python-binary-name",
        type=str,
        help="Python binary name to use to install requirements",
        default="python3",
    )
    build_parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help=(
            "[experimental] show detailed information about "
            "created/copied/modified/conflict files after build is complete"
        ),
    )
    build_parser.add_argument(
        "--pip-version",
        type=str,
        help="pip version that will be used to install libraries.",
        default="latest",
    )
    build_parser.add_argument(
        "--pip-legacy-resolver",
        action="store_true",
        default=False,
        help="Use old pip dependency resolver by adding flag '--use-deprecated=legacy-resolver' "
        "to pip install command.",
    )
    build_parser.add_argument(
        "--pip-custom-flag",
        help="Custom flag that will be add to pip install command",
        type=str,
        default=False,
        required=False,
    )
    build_parser.add_argument(
        "--build-custom-ui",
        help="Custom flag decides whether to try building custom UI code",
        action="store_true",
        default=False,
        required=False,
    )

    package_parser = subparsers.add_parser("package", description="Package an add-on")
    package_parser.add_argument(
        "--path",
        required=True,
        type=str,
        help="path to the built add-on (should include app.manifest file)",
    )
    package_parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="output path to store archived add-on",
        default=None,
    )

    init_parser = subparsers.add_parser(
        "init", description="Initialize an empty add-on"
    )
    init_parser.add_argument(
        "--addon-name",
        type=str,
        help="add-on name",
        required=True,
    )
    init_parser.add_argument(
        "--addon-rest-root",
        type=str,
        help="add-on REST root",
        required=False,
        default=None,
    )
    init_parser.add_argument(
        "--addon-display-name",
        type=str,
        help="add-on display name",
        required=True,
    )
    init_parser.add_argument(
        "--addon-input-name",
        type=str,
        help="add-on input name",
        required=True,
    )
    init_parser.add_argument(
        "--addon-version",
        type=str,
        help="add-on version",
        default="0.0.1",
    )
    init_parser.add_argument(
        "--overwrite",
        action="store_true",
        default=False,
        help="overwrite already generated add-on folder",
    )
    init_parser.add_argument(
        "--need-proxy",
        action="store_true",
        required=False,
        help="Specifies if proxy is needed or not",
    )
    init_parser.add_argument(
        "--add-license",
        type=str,
        choices=[
            "Apache License 2.0",
            "MIT License",
            "SPLUNK PRE-RELEASE SOFTWARE LICENSE AGREEMENT",
        ],
        help=(
            "adds any one of license agreement such as 'Apache License 2.0', 'MIT License', or "
            "'SPLUNK PRE-RELEASE SOFTWARE LICENSE AGREEMENT' to the `<your_add-on_name>/package/LICENSES` directory."
        ),
        required=False,
        default=None,
    )
    init_parser.add_argument(
        "--include-author",
        type=str,
        help="adds author in app.mainifest under `info -> author -> name` field",
        required=False,
        default=None,
    )

    import_from_aob_parser = subparsers.add_parser(
        "import-from-aob", description="[Experimental] Import from AoB"
    )
    import_from_aob_parser.add_argument(
        "--addon-name",
        type=str,
        help="add-on name",
        required=True,
    )
    validate_parser = subparsers.add_parser(
        "validate", description="Used to validate Splunk app using Splunk appinspect"
    )
    validate_parser.add_argument(
        "--addon-path",
        type=str,
        help="Path of the Splunk app.",
        required=True,
    )

    publish_parser = subparsers.add_parser(
        "publish", description="Publish package to the Splunkbase"
    )
    publish_parser.add_argument(
        "--app-id",
        type=int,
        help="Splunkbase numerical app id listed in the URL of the app details page.",
        required=True,
    )
    publish_parser.add_argument(
        "--package-path",
        type=str,
        help="path to the package file (.tar.gz) to be published",
        required=True,
    )
    publish_parser.add_argument(
        "--splunk-versions",
        type=str,
        help="The Splunk version(s) that the package is compatible with. For example, '9.1,9.2'.",
        required=True,
    )
    publish_parser.add_argument(
        "--cim-versions",
        type=str,
        help="The CIM version(s) that the release is compatible with. For example, '4.9,6.1'.",
        required=True,
    )
    publish_parser.add_argument(
        "--make-visible",
        dest="visibility",
        help="The release is to be visible upon package validation success.",
        action="store_true",
    )
    publish_parser.add_argument(
        "--username", type=str, help="Username of the splunkbase account", required=True
    )
    publish_parser.add_argument(
        "--password", type=str, help="Password of the splunkbase account", required=True
    )

    args = parser.parse_args(argv)
    if args.command == "build":
        build.generate(
            source=args.source,
            config_path=args.config,
            addon_version=args.ta_version,
            output_directory=args.output,
            python_binary_name=args.python_binary_name,
            verbose_file_summary_report=args.verbose,
            pip_version=args.pip_version,
            pip_legacy_resolver=args.pip_legacy_resolver,
            pip_custom_flag=args.pip_custom_flag,
            build_custom_ui=args.build_custom_ui,
        )
    if args.command == "package":
        package.package(path_to_built_addon=args.path, output_directory=args.output)
    if args.command == "validate":
        validate.validate(file_path=args.addon_path)
    if args.command == "init":
        init.init(
            addon_name=args.addon_name,
            addon_rest_root=args.addon_rest_root,
            addon_display_name=args.addon_display_name,
            addon_input_name=args.addon_input_name,
            addon_version=args.addon_version,
            overwrite=args.overwrite,
            need_proxy=args.need_proxy,
            add_license=args.add_license,
            include_author=args.include_author,
        )
    if args.command == "import-from-aob":
        import_from_aob.import_from_aob(
            addon_name=args.addon_name,
        )
    if args.command == "publish":
        publish.publish_package(
            app_id=args.app_id,
            package_path=args.package_path,
            splunk_versions=args.splunk_versions,
            cim_versions=args.cim_versions,
            visibility=args.visibility,
            username=args.username,
            password=args.password,
        )
    return 0


def config_file_type(filename: str) -> str:
    pattern = re.compile(r".*\.(json|yaml)$")
    if not pattern.match(filename):
        msg = f"Global config file should be a JSON or YAML file. Provided: {filename}"
        logger.error(msg)
        raise argparse.ArgumentTypeError(msg)
    return filename


if __name__ == "__main__":
    raise SystemExit(main())



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from ..file_generator import FileGenerator
from .create_alert_actions_conf import AlertActionsConf
from .create_app_conf import AppConf
from .create_eventtypes_conf import EventtypesConf
from .create_inputs_conf import InputsConf
from .create_restmap_conf import RestMapConf
from .create_server_conf import ServerConf
from .create_tags_conf import TagsConf
from .create_web_conf import WebConf
from .create_account_conf import AccountConf
from .create_settings_conf import SettingsConf
from .create_commands_conf import CommandsConf
from .create_searchbnf_conf import SearchbnfConf

__all__ = [
    "FileGenerator",
    "ServerConf",
    "RestMapConf",
    "WebConf",
    "AlertActionsConf",
    "EventtypesConf",
    "TagsConf",
    "AppConf",
    "InputsConf",
    "AccountConf",
    "SettingsConf",
    "CommandsConf",
    "SearchbnfConf",
]



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, Optional
from collections.abc import Iterable

from splunk_add_on_ucc_framework.commands.openapi_generator import oas


_EAI_OUTPUT_MODE = {
    "name": "output_mode",
    "in": "query",
    "required": True,
    "description": "Output mode",
    "schema": {
        "type": "string",
        "enum": ["json"],
        "default": "json",
    },
}
EAI_DEFAULT_PARAMETERS = [_EAI_OUTPUT_MODE]
EAI_DEFAULT_PARAMETERS_SPECIFIED = [
    _EAI_OUTPUT_MODE,
    {
        "name": "name",
        "in": "path",
        "required": True,
        "description": "The name of the item to operate on",
        "schema": {"type": "string"},
    },
]


def _eai_response_schema(schema: Any) -> oas.MediaTypeObject:
    return oas.MediaTypeObject(
        schema={
            "type": "object",
            "properties": {
                "entry": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "name": {"type": "string"},
                            "content": schema,
                        },
                    },
                }
            },
        }
    )


@dataclass
class EndpointRegistrationEntry:
    """
    Represents an entry in the endpoint registration file.
    """

    name: str
    rh_name: str
    actions_list: list[str]
    capabilities: dict[str, str] = field(default_factory=dict)
    resourcePresent: bool = False
    parentResource: str = "/"

    def actions(self) -> list[str]:
        """
        Method for consistency with RestEndpointBuilder.
        """
        return self.actions_list


@dataclass
class RestHandlerConfig:
    """
    Represents a REST handler configuration. See schema.json.
    """

    name: str
    endpoint: str
    handlerType: str
    capabilities: dict[str, str] = field(default_factory=dict)
    resourcePresent: bool = False
    parentResource: str = "/"
    registerHandler: Optional[dict[str, Any]] = None
    requestParameters: Optional[dict[str, dict[str, Any]]] = None
    responseParameters: Optional[dict[str, dict[str, Any]]] = None

    @property
    def request_parameters(self) -> dict[str, dict[str, Any]]:
        return self.requestParameters or {}

    @property
    def response_parameters(self) -> dict[str, dict[str, Any]]:
        return self.responseParameters or {}

    @property
    def supported_actions(self) -> set[str]:
        actions = set((self.registerHandler or {}).get("actions", []))
        actions.update(self.request_parameters.keys())
        actions.update(self.response_parameters.keys())
        return actions

    def _eai_params_to_schema_object(
        self, params: Optional[dict[str, Any]]
    ) -> Optional[dict[str, Any]]:
        if not params:
            return None

        obj: dict[str, Any] = {
            "type": "object",
            "properties": {},
        }

        required = []

        for name, param in params.items():
            obj["properties"][name] = param["schema"]

            if param.get("required", False):
                required.append(name)

        if required:
            obj["required"] = required

        return obj

    def _oas_object_eai_list_or_remove(
        self, description: str, action: str
    ) -> Optional[oas.OperationObject]:
        if action not in self.supported_actions:
            return None

        op_obj: dict[str, Any] = {
            "description": description,
            "responses": {
                "200": oas.ResponseObject(description=description),
            },
        }

        if self.request_parameters.get(action):
            op_obj["parameters"] = [
                {
                    "name": key,
                    "in": "query",
                    "required": item.get("required", False),
                    "schema": item["schema"],
                }
                for key, item in self.request_parameters[action].items()
            ]

        if self.response_parameters.get(action):
            op_obj["responses"]["200"].content = {
                "application/json": _eai_response_schema(
                    self._eai_params_to_schema_object(self.response_parameters[action])
                )
            }

        return oas.OperationObject(**op_obj)

    def _oas_object_eai_create_or_edit(
        self, description: str, action: str
    ) -> Optional[oas.OperationObject]:
        if action not in self.supported_actions:
            return None

        request_parameters = deepcopy(self.request_parameters.get(action, {}))

        if action == "create":
            request_parameters["name"] = {
                "schema": {"type": "string"},
                "required": True,
            }

        op_obj: dict[str, Any] = {
            "description": description,
            "responses": {
                "200": oas.ResponseObject(description=description),
            },
        }

        if request_parameters:
            op_obj["requestBody"] = oas.RequestBodyObject(
                content={
                    "application/x-www-form-urlencoded": {
                        "schema": self._eai_params_to_schema_object(request_parameters),
                    },
                }
            )

        if self.response_parameters.get(action):
            op_obj["responses"]["200"].content = {
                "application/json": _eai_response_schema(
                    self._eai_params_to_schema_object(self.response_parameters[action])
                )
            }

        return oas.OperationObject(**op_obj)

    def _oas_object_eai_list_all(self) -> Optional[oas.OperationObject]:
        return self._oas_object_eai_list_or_remove(
            f"Get list of items for {self.name}", "list"
        )

    def _oas_object_eai_list_one(self) -> Optional[oas.OperationObject]:
        return self._oas_object_eai_list_or_remove(
            f"Get {self.name} item details", "list"
        )

    def _oas_object_eai_create(self) -> Optional[oas.OperationObject]:
        return self._oas_object_eai_create_or_edit(
            f"Create item in {self.name}", "create"
        )

    def _oas_object_eai_edit(self) -> Optional[oas.OperationObject]:
        return self._oas_object_eai_create_or_edit(f"Update {self.name} item", "edit")

    def _oas_object_eai_remove(self) -> Optional[oas.OperationObject]:
        return self._oas_object_eai_list_or_remove(f"Delete {self.name} item", "remove")

    def _oas_objects_eai_normal(self) -> dict[str, oas.PathItemObject]:
        endpoint = self.endpoint.strip("/")

        obj: dict[str, Any] = {}
        list_all = self._oas_object_eai_list_all()

        if list_all:
            obj["get"] = list_all

        create = self._oas_object_eai_create()

        if create:
            obj["post"] = create

        if obj:
            obj["parameters"] = EAI_DEFAULT_PARAMETERS
            return {f"/{endpoint}": oas.PathItemObject(**obj)}

        return {}

    def _oas_objects_eai_specified(self) -> dict[str, oas.PathItemObject]:
        endpoint = self.endpoint.strip("/")

        obj_specified: dict[str, Any] = {}

        list_one = self._oas_object_eai_list_one()

        if list_one:
            obj_specified["get"] = list_one

        edit = self._oas_object_eai_edit()

        if edit:
            obj_specified["post"] = edit

        remove = self._oas_object_eai_remove()

        if remove:
            obj_specified["delete"] = remove

        if obj_specified:
            obj_specified["parameters"] = EAI_DEFAULT_PARAMETERS_SPECIFIED
            return {f"/{endpoint}/{{name}}": oas.PathItemObject(**obj_specified)}

        return {}

    def _oas_objects_eai(self) -> dict[str, oas.PathItemObject]:
        obj_dict: dict[str, oas.PathItemObject] = {}

        obj_dict.update(self._oas_objects_eai_normal())
        obj_dict.update(self._oas_objects_eai_specified())

        return obj_dict

    @property
    def oas_paths(self) -> dict[str, oas.PathItemObject]:
        if self.handlerType == "EAI":
            return self._oas_objects_eai()
        raise ValueError(f"Unsupported handler type: {self.handlerType}")

    @property
    def endpoint_registration_entry(self) -> Optional[EndpointRegistrationEntry]:
        if not self.registerHandler:
            return None

        if not self.registerHandler.get("actions") or not self.registerHandler.get(
            "file"
        ):
            return None

        file: str = self.registerHandler["file"]

        if file.endswith(".py"):
            file = file[:-3]

        return EndpointRegistrationEntry(
            name=self.endpoint,
            rh_name=file,
            actions_list=self.registerHandler["actions"],
            resourcePresent=self.resourcePresent,
            parentResource=self.parentResource,
            capabilities=self.capabilities,
        )


class UserDefinedRestHandlers:
    """
    Represents a logic for dealing with user-defined REST handlers
    """

    def __init__(self, definitions: Iterable[dict[str, Any]]) -> None:
        self._definitions: list[RestHandlerConfig] = []
        self._names: set[str] = set()
        self._endpoints: set[str] = set()

        for definition in definitions:
            rest_handler_config_definition = RestHandlerConfig(**definition)
            if rest_handler_config_definition.name in self._names:
                raise ValueError(
                    f"REST handler defined in Global Config contains duplicated name: "
                    f"{rest_handler_config_definition.name}. "
                    "Please change it to a unique name."
                )

            if rest_handler_config_definition.endpoint in self._endpoints:
                raise ValueError(
                    f"REST handler defined in Global Config contains duplicated endpoint: "
                    f"{rest_handler_config_definition.endpoint} "
                    f"(name={rest_handler_config_definition.name}). Please change it to a unique endpoint."
                )

            self._names.add(rest_handler_config_definition.name)
            self._endpoints.add(rest_handler_config_definition.endpoint)

            self._definitions.append(rest_handler_config_definition)

    @property
    def oas_paths(self) -> dict[str, oas.PathItemObject]:
        paths = {}

        for definition in self._definitions:
            paths.update(definition.oas_paths)

        return paths

    @property
    def endpoint_registration_entries(self) -> list[EndpointRegistrationEntry]:
        entries = []

        for definition in self._definitions:
            entry = definition.endpoint_registration_entry

            if entry is None:
                continue

            entries.append(entry)

        return entries



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import NamedTuple
from .file_generator import FileGenerator

from splunk_add_on_ucc_framework.generators.xml_files import (
    ConfigurationXml,
    DashboardXml,
    DefaultXml,
    InputsXml,
    RedirectXml,
)
from splunk_add_on_ucc_framework.generators.html_files import AlertActionsHtml
from splunk_add_on_ucc_framework.generators.python_files import CustomCommandPy
from splunk_add_on_ucc_framework.generators.conf_files import (
    AlertActionsConf,
    AppConf,
    EventtypesConf,
    InputsConf,
    RestMapConf,
    ServerConf,
    TagsConf,
    WebConf,
    AccountConf,
    SettingsConf,
    CommandsConf,
    SearchbnfConf,
)

__all__ = ["FileClass", "GEN_FILE_LIST"]


class FileClass(NamedTuple):
    file_name: str
    file_class: type[FileGenerator]
    file_path: list[str]


GEN_FILE_LIST: list[FileClass] = [
    FileClass("app.conf", AppConf, ["default"]),
    FileClass("inputs.conf", InputsConf, ["default"]),
    FileClass("server.conf", ServerConf, ["default"]),
    FileClass("restmap.conf", RestMapConf, ["default"]),
    FileClass("web.conf", WebConf, ["default"]),
    FileClass(
        "alert_actions.conf",
        AlertActionsConf,
        ["default"],
    ),
    FileClass("eventtypes.conf", EventtypesConf, ["default"]),
    FileClass("tags.conf", TagsConf, ["default"]),
    FileClass("commands.conf", CommandsConf, ["default"]),
    FileClass("searchbnf.conf", SearchbnfConf, ["default"]),
    FileClass("_account.conf", AccountConf, ["README"]),
    FileClass("_settings.conf", SettingsConf, ["README"]),
    FileClass(
        "configuration.xml",
        ConfigurationXml,
        ["default", "data", "ui", "views"],
    ),
    FileClass(
        "dashboard.xml",
        DashboardXml,
        ["default", "data", "ui", "views"],
    ),
    FileClass(
        "default.xml",
        DefaultXml,
        ["default", "data", "ui", "nav"],
    ),
    FileClass(
        "inputs.xml",
        InputsXml,
        ["default", "data", "ui", "views"],
    ),
    FileClass(
        "_redirect.xml",
        RedirectXml,
        ["default", "data", "ui", "views"],
    ),
    FileClass(
        "_.html",
        AlertActionsHtml,
        ["default", "data", "ui", "alerts"],
    ),
    FileClass(
        "_.py",
        CustomCommandPy,
        ["bin"],
    ),
]



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import os
import re
from typing import Any
import logging
import itertools
from splunk_add_on_ucc_framework.const import SPLUNK_COMMANDS

import jsonschema

from splunk_add_on_ucc_framework import dashboard as dashboard_lib
from splunk_add_on_ucc_framework import global_config as global_config_lib
from splunk_add_on_ucc_framework.tabs import Tab
from splunk_add_on_ucc_framework.exceptions import GlobalConfigValidatorException

logger = logging.getLogger("ucc_gen")

# The entity types that do not allow to add validators (so the warning will not appear)
ENTITY_TYPES_WITHOUT_VALIDATORS = {
    "radio",
    "index",
    "helpLink",
    "checkbox",
    "interval",
    "custom",
}


class GlobalConfigValidator:
    """
    GlobalConfigValidator implements different validation for globalConfig file.
    Custom validation should be implemented here.
    """

    def __init__(
        self,
        internal_root_dir: str,
        global_config: global_config_lib.GlobalConfig,
        source: str = "",
    ):
        self._internal_root_dir = internal_root_dir
        self._source_dir = source
        self._global_config = global_config
        self._config = global_config.content
        self.resolved_configuration = global_config.resolved_configuration

    def _validate_config_against_schema(self) -> None:
        """
        Validates config against JSON schema.
        Raises jsonschema.ValidationError if config is not valid.
        """
        schema_path = os.path.join(self._internal_root_dir, "schema", "schema.json")
        with open(schema_path, encoding="utf-8") as f_schema:
            schema_raw = f_schema.read()
            schema = json.loads(schema_raw)
        try:
            return jsonschema.validate(instance=self._config, schema=schema)
        except jsonschema.ValidationError as e:
            raise GlobalConfigValidatorException(e.message)

    def _validate_configuration_tab_table_has_name_field(self) -> None:
        """
        Validates that if a configuration tab should be rendered as a table,
        then it needs to have an entity which has field "name".
        """
        for tab in self.resolved_configuration:
            if "table" in tab:
                entities = tab.get("entity", [])
                has_name_field = False
                for entity in entities:
                    if entity["field"] == "name":
                        has_name_field = True
                        break
                if not has_name_field:
                    raise GlobalConfigValidatorException(
                        f"Tab '{tab['name']}' should have entity with field 'name'"
                    )

    def _validate_custom_rest_handlers(self) -> None:
        """
        Validates that only "restHandlerName" or both "restHandlerModule" and
        "restHandlerClass" is present in the input configuration. Also validates
        that both "restHandlerModule" and "restHandlerClass" is present if any
        of them are present.

        The valid scenarios:
            * only restHandlerName is present
            * both restHandlerModule and restHandlerClass is present
        Everything other combination is considered invalid.
        """
        pages = self._config["pages"]
        inputs = pages.get("inputs")
        if inputs is None:
            return
        services = inputs["services"]
        for service in services:
            rest_handler_name = service.get("restHandlerName")
            rest_handler_module = service.get("restHandlerModule")
            rest_handler_class = service.get("restHandlerClass")
            if rest_handler_name is not None and (
                rest_handler_module is not None or rest_handler_class is not None
            ):
                raise GlobalConfigValidatorException(
                    f"Input '{service['name']}' has both 'restHandlerName' and "
                    f"'restHandlerModule' or 'restHandlerClass' fields present. "
                    f"Please use only 'restHandlerName' or 'restHandlerModule' "
                    f"and 'restHandlerClass'."
                )
            if (rest_handler_module is not None and rest_handler_class is None) or (
                rest_handler_module is None and rest_handler_class is not None
            ):
                raise GlobalConfigValidatorException(
                    f"Input '{service['name']}' should have both "
                    f"'restHandlerModule' and 'restHandlerClass' fields "
                    f"present, only 1 of them was found."
                )

    def _validate_file_type_entity(self) -> None:
        """
        Validates that file-based field has all necessary fields.
        Things should be provided in case of file field:
            * options field
            * supportedFileTypes field should be present in options
        Also if file is encrypted but not required, this is not supported,
        and we need to throw a validation error.
        """
        for tab in self.resolved_configuration:
            # For customTab entity is optional
            entities = tab.get("entity", [])
            for entity in entities:
                if entity["type"] == "file":
                    is_required = entity.get("required", False)
                    is_encrypted = entity.get("encrypted", False)
                    if is_encrypted and not is_required:
                        raise GlobalConfigValidatorException(
                            f"Field {entity['field']} uses type 'file' which is encrypted and not required, "
                            f"this is not supported"
                        )
                    options = entity.get("options")
                    if options is None:
                        raise GlobalConfigValidatorException(
                            f"Options field for the file type should be present "
                            f"for '{entity['field']}' field."
                        )
                    supported_file_types = options.get("supportedFileTypes")
                    if supported_file_types is None:
                        raise GlobalConfigValidatorException(
                            f"You should define your supported file types in "
                            f"the `supportedFileTypes` field for the "
                            f"'{entity['field']}' field."
                        )

    def _validate_string_validator(
        self, entity_field: str, validator: dict[str, Any]
    ) -> None:
        """
        Validates string validator. maxLength should be greater or equal
        than minLength.
        """
        if validator["maxLength"] < validator["minLength"]:
            raise GlobalConfigValidatorException(
                f"Entity '{entity_field}' has incorrect string validator, "
                f"'maxLength' should be greater or equal than 'minLength'."
            )

    def _validate_number_validator(
        self, entity_field: str, validator: dict[str, Any]
    ) -> None:
        """
        Validates number validator, both values in range should be numbers and
        first one should be smaller than the second one.
        """
        validator_range = validator["range"]
        if len(validator_range) != 2:
            raise GlobalConfigValidatorException(
                f"Entity '{entity_field}' has incorrect number validator, "
                f"it should have 2 elements under 'range' field."
            )
        # Schema already checks that the values in the 'range' field are
        # numbers.
        if validator_range[1] < validator_range[0]:
            raise GlobalConfigValidatorException(
                f"Entity '{entity_field}' has incorrect number validator, "
                f"second element should be greater or equal than first element."
            )

    def _validate_regex_validator(
        self, entity_field: str, validator: dict[str, Any]
    ) -> None:
        """
        Validates regex validator, provided regex should at least be compilable.
        """
        try:
            re.compile(validator["pattern"])
        except re.error:
            raise GlobalConfigValidatorException(
                f"Entity '{entity_field}' has incorrect regex validator, "
                f"pattern provided in the 'pattern' field is not compilable."
            )

    def _validate_interval(self, entity: dict[str, Any]) -> None:
        """
        Validate options of the 'interval' entity
        """
        if "range" in entity.get("options", {}):
            self._validate_number_validator(entity["field"], entity["options"])

    def _validate_entity_validators(self, entity: dict[str, Any]) -> None:
        """
        Validates entity validators.
        """
        validators = entity.get("validators", [])

        if should_warn_on_empty_validators(entity):
            logger.warning(
                f"The field '{entity.get('field')}' does not have a validator specified. It's recommended "
                "to add a validator to ensure the security and integrity of the input data. For more "
                "information, please refer to the documentation."
            )

        for validator in validators:
            if validator["type"] == "string":
                self._validate_string_validator(entity["field"], validator)
            if validator["type"] == "number":
                self._validate_number_validator(entity["field"], validator)
            if validator["type"] == "regex":
                self._validate_regex_validator(entity["field"], validator)

        if entity.get("type", "") == "interval":
            self._validate_interval(entity)

    def _validate_validators(self) -> None:
        """
        Validates both configuration and services validators, currently string,
        number and regex are supported.
        """
        pages = self._config["pages"]
        for tab in self.resolved_configuration:
            entities = tab.get("entity", [])
            for entity in entities:
                self._validate_entity_validators(entity)

        inputs = pages.get("inputs")
        if inputs is None:
            return
        services = inputs["services"]
        for service in services:
            entities = service["entity"]
            for entity in entities:
                self._validate_entity_validators(entity)

    @staticmethod
    def _find_duplicates_in_list(_list: list[Any]) -> bool:
        return len(set(_list)) != len(_list)

    def _validate_children_duplicates(
        self, children: list[dict[str, Any]], entity_label: str
    ) -> None:
        """
        Validates duplicates under children key in autoCompleteFields
        for fields under keys: label, value
        """
        labels, values = [], []
        for child in children:
            labels.append(child["label"].lower())
            child_value = child["value"]
            if isinstance(child_value, str):
                values.append(child_value.lower())
            else:
                values.append(child_value)
        if self._find_duplicates_in_list(values) or self._find_duplicates_in_list(
            labels
        ):
            raise GlobalConfigValidatorException(
                f"Duplicates found for autoCompleteFields children in entity '{entity_label}'"
            )

    def _validate_autoCompleteFields_duplicates(
        self, options: dict[str, Any], entity_label: str
    ) -> None:
        """
        Validates duplicates in autoCompleteFields keys
        for fields under keys: label, value
        If autoCompleteFields has children key, children validator is called
        """
        labels, values = [], []
        for field in options["autoCompleteFields"]:
            labels.append(field.get("label").lower())
            children = field.get("children")
            if children:
                self._validate_children_duplicates(children, entity_label)
            else:
                field_value = field.get("value")
                if isinstance(field_value, str):
                    values.append(field_value.lower())
                else:
                    values.append(field_value)
        if self._find_duplicates_in_list(values) or self._find_duplicates_in_list(
            labels
        ):
            raise GlobalConfigValidatorException(
                f"Duplicates found for autoCompleteFields: '{entity_label}'"
            )

    def _validate_multilevel_menu(self) -> None:
        inputs = self._config.get("pages").get("inputs")
        if inputs:
            groups_menu = inputs.get("groupsMenu")
            if groups_menu:
                names, titles = [], []
                for group in groups_menu:
                    names.append(group.get("groupName").lower())
                    titles.append(group.get("groupTitle").lower())
                    group_services = group.get("groupServices")
                    if group_services:
                        for serviceName in group_services:
                            if not any(
                                serviceName == service.get("name")
                                for service in inputs.get("services")
                            ):
                                raise GlobalConfigValidatorException(
                                    f"{serviceName} ServiceName in the "
                                    f"multi-level menu does not match any "
                                    f"services name."
                                )
                    else:
                        if not any(
                            group.get("groupName") == service.get("name")
                            and group.get("groupTitle") == service.get("title")
                            for service in inputs.get("services")
                        ):
                            raise GlobalConfigValidatorException(
                                f'{group.get("groupName")} groupName or '
                                f'{group.get("groupTitle")} groupTitle in the '
                                f"multi-level menu does not match any services "
                                f"name or title."
                            )

                if self._find_duplicates_in_list(
                    names
                ) or self._find_duplicates_in_list(titles):
                    raise GlobalConfigValidatorException(
                        "Duplicates found for multi-level menu groups' names or titles."
                    )

    def _validate_entity_duplicates(self, entity: list[dict[str, Any]]) -> None:
        """
        Validates duplicates in entity keys
        for fields under keys: field, label
        If entity has autoCompleteFields key, autoCompleteFields validator is called
        """
        fields, labels = [], []
        for _entity in entity:
            fields.append(_entity["field"].lower())
            if "label" in _entity:
                labels.append(_entity["label"].lower())
            options = _entity.get("options")
            if options and options.get("autoCompleteFields"):
                self._validate_autoCompleteFields_duplicates(
                    _entity["options"], _entity["label"]
                )
        if self._find_duplicates_in_list(fields) or self._find_duplicates_in_list(
            labels
        ):
            raise GlobalConfigValidatorException(
                "Duplicates found for entity field or label"
            )

    def _validate_tabs_duplicates(self, tabs: list[Tab]) -> None:
        """
        Validates duplicates in tab keys under configuration
        for fields under keys: name, title
        Calls for entity validator, as at least one entity is required in schema
        """
        names, titles, types = [], [], []
        for tab in tabs:
            names.append(tab["name"].lower())
            titles.append(tab["title"].lower())

            if tab.tab_type is not None:
                types.append(tab.tab_type.lower())

            self._validate_entity_duplicates(tab.get("entity", []))
        if (
            self._find_duplicates_in_list(names)
            or self._find_duplicates_in_list(titles)
            or self._find_duplicates_in_list(types)
        ):
            raise GlobalConfigValidatorException(
                "Duplicates found for tabs names, titles or types"
            )

    def _validate_inputs_duplicates(self, inputs: dict[str, Any]) -> None:
        """
        Validates duplicates in tab keys under configuration
        for fields under keys: name, title
        Inputs are not required in schema
        """
        names, titles = [], []
        for service in inputs["services"]:
            names.append(service["name"].lower())
            titles.append(service["title"].lower())

            self._validate_entity_duplicates(service["entity"])

        if self._find_duplicates_in_list(names) or self._find_duplicates_in_list(
            titles
        ):
            raise GlobalConfigValidatorException(
                "Duplicates found for inputs (services) names or titles"
            )

    def _validate_duplicates(self) -> None:
        """
        Validates duplicates for both tabs and services (inputs). Inputs however are
        not required in schema, so this checks if globalConfig has inputs
        """
        pages = self._config["pages"]
        self._validate_tabs_duplicates(self.resolved_configuration)

        inputs = pages.get("inputs")
        if inputs:
            self._validate_inputs_duplicates(inputs)

    def _validate_alerts(self) -> None:
        # TODO: test cases should be added here.
        alerts = self._config.get("alerts", [])
        for alert in alerts:
            fields = []
            alert_entity = alert.get("entity")
            if alert_entity is None:
                continue
            for entity in alert_entity:
                if entity.get("field") in fields:
                    raise GlobalConfigValidatorException(
                        "Field names should be unique across alerts"
                    )
                else:
                    fields.append(entity.get("field"))

    def _validate_panels(self) -> None:
        """
        Validates if the panels defined in the configuration are supported.
        """
        dashboard = self._config["pages"].get("dashboard")
        if dashboard:
            for panel in dashboard["panels"]:
                if panel["name"] not in dashboard_lib.SUPPORTED_PANEL_NAMES:
                    raise GlobalConfigValidatorException(
                        f"'{panel['name']}' is not a supported panel name. "
                        f"Supported panel names: {dashboard_lib.SUPPORTED_PANEL_NAMES_READABLE}"
                    )

    def _validate_checkbox_group(self) -> None:
        pages = self._config["pages"]
        inputs = pages.get("inputs")
        if inputs is None:
            return
        services = inputs["services"]
        for service in services:
            for entity in service["entity"]:
                if (
                    entity["type"] == "checkboxGroup"
                    or entity["type"] == "checkboxTree"
                ):
                    row_field_names = []
                    for row in entity["options"]["rows"]:
                        if row["field"] in row_field_names:
                            raise GlobalConfigValidatorException(
                                f"Entity {entity['field']} has duplicate field ({row['field']}) in options.rows"
                            )
                        row_field_names.append(row["field"])
                    groups = entity["options"].get("groups")
                    if groups is None:
                        return
                    group_used_field_names = []
                    for group in groups:
                        for group_field_name in group["fields"]:
                            if group_field_name not in row_field_names:
                                raise GlobalConfigValidatorException(
                                    f"Entity {entity['field']} uses field ({group_field_name}) "
                                    f"which is not defined in options.rows"
                                )
                            if group_field_name in group_used_field_names:
                                raise GlobalConfigValidatorException(
                                    f"Entity {entity['field']} has duplicate field ({group_field_name}) "
                                    f"in options.groups"
                                )
                            group_used_field_names.append(group_field_name)

    def _validate_groups(self) -> None:
        pages = self._config["pages"]
        inputs = pages.get("inputs")
        if inputs is None:
            return
        services = inputs["services"]
        for service in services:
            groups = service.get("groups")
            if groups is None:
                continue
            entities = service["entity"]
            entity_fields = []
            for entity in entities:
                entity_fields.append(entity["field"])
            service_group_labels = []
            for group in groups:
                group_label = group["label"]
                if group_label in service_group_labels:
                    raise GlobalConfigValidatorException(
                        f"Service {service['name']} has duplicate labels in groups"
                    )
                service_group_labels.append(group_label)
            for group in groups:
                group_fields = group["fields"]
                for group_field in group_fields:
                    if group_field not in entity_fields:
                        raise GlobalConfigValidatorException(
                            f"Service {service['name']} uses group field {group_field} which is not defined in entity"
                        )

    def _is_circular_modification(
        self,
        mods: list[Any],
        visited: dict[str, str],
        all_entity_fields: list[Any],
        current_field: str,
    ) -> dict[str, str]:
        """
        Checks if there is circular modification based on visited list and DFS algorithm
        """
        DEAD_END = "dead_end"
        VISITING = "visited"
        visited[current_field] = VISITING

        current_field_mods = next(
            (mod for mod in mods if mod["fieldId"] == current_field), None
        )
        if current_field_mods is None:
            # no more dependent modification fields
            visited[current_field] = DEAD_END
            return visited

        if current_field in current_field_mods["influenced_fields_value_change"]:
            # field can modify itself except "value" property
            raise GlobalConfigValidatorException(
                f"""Field '{current_field}' tries to modify itself value"""
            )

        for influenced_field in current_field_mods["influenced_fields"]:
            if influenced_field not in all_entity_fields:
                raise GlobalConfigValidatorException(
                    f"""Modification in field '{current_field}' for not existing field '{influenced_field}'"""
                )

            if influenced_field in current_field_mods["influenced_fields_value_change"]:
                if visited[influenced_field] == VISITING:
                    raise GlobalConfigValidatorException(
                        f"""Circular modifications for field '{influenced_field}' in field '{current_field}'"""
                    )
                # check next influenced by value change field
                visited = self._is_circular_modification(
                    mods, visited, all_entity_fields, influenced_field
                )

        # All dependent modifications fields are dead_end
        visited[current_field] = DEAD_END
        return visited

    def _check_if_circular_modification(
        self,
        all_entity_fields: list[Any],
        fields_with_mods: list[Any],
        modifications: list[Any],
    ) -> None:
        visited = {field: "not_visited" for field in all_entity_fields}
        for start_field in fields_with_mods:
            # DFS algorithm for all fields with modifications
            visited = self._is_circular_modification(
                modifications, visited, all_entity_fields, start_field
            )

    @staticmethod
    def _get_mods_data_for_single_entity(
        entity: dict[str, Any],
    ) -> list[Any]:
        """
        Get modification entity data as lists
        """
        entity_modifications = []
        if "modifyFieldsOnValue" in entity:
            influenced_fields_value_change = set()
            influenced_fields = set()
            for mods in entity["modifyFieldsOnValue"]:
                for mod in mods["fieldsToModify"]:
                    influenced_fields.add(mod["fieldId"])

                    if (
                        mod.get("value") is not None
                    ):  # circular deps are not a problem if not about value
                        influenced_fields_value_change.add(mod["fieldId"])
            entity_modifications.append(
                {
                    "fieldId": entity["field"],
                    "influenced_fields": influenced_fields,
                    "influenced_fields_value_change": influenced_fields_value_change,
                }
            )
        return entity_modifications

    @staticmethod
    def _get_all_entities(
        collections: list[dict[str, Any]],
    ) -> list[Any]:
        all_fields = []

        tab_entities: list[Any] = [
            el.get("entity") for el in collections if el.get("entity")
        ]
        all_entities = list(itertools.chain.from_iterable(tab_entities))

        for entity in all_entities:
            if entity["type"] == "oauth":
                for oauthType in entity["options"]["auth_type"]:
                    for oauthEntity in entity["options"][oauthType]:
                        all_fields.append(oauthEntity)
            else:
                all_fields.append(entity)

        return all_fields

    def _get_all_modification_data(
        self,
        collections: list[dict[str, Any]],
    ) -> list[Any]:
        fields_with_mods: list[Any] = []
        all_modifications: list[Any] = []
        all_fields: list[str] = []

        entities = self._get_all_entities(collections)
        for entity in entities:
            all_fields.append(entity["field"])

            if "modifyFieldsOnValue" in entity:
                fields_with_mods.append(entity["field"])
                entity_mods = self._get_mods_data_for_single_entity(entity)
                all_modifications.extend(entity_mods)

        return [fields_with_mods, all_modifications, all_fields]

    def _validate_field_modifications(self) -> None:
        """
        Validates:
        * Circular dependencies
        * If fields try to modify itself
        * If fields try to modify field that do not exist
        """
        pages = self._config["pages"]

        if "configuration" in pages:
            configuration = pages["configuration"]
            tabs = configuration["tabs"]

            (
                fields_with_mods_config,
                all_modifications_config,
                all_fields_config,
            ) = self._get_all_modification_data(tabs)

            self._check_if_circular_modification(
                all_fields_config, fields_with_mods_config, all_modifications_config
            )

        if "inputs" in pages:
            inputs = pages["inputs"]
            services = inputs["services"]

            (
                fields_with_mods_inputs,
                all_modifications_inputs,
                all_fields_inputs,
            ) = self._get_all_modification_data(services)

            self._check_if_circular_modification(
                all_fields_inputs, fields_with_mods_inputs, all_modifications_inputs
            )

    def _validate_meta_default_view(self) -> None:
        default_view = self._global_config.meta.get("defaultView")
        if (
            default_view == "configuration"
            and not self._global_config.has_configuration()
        ):
            raise GlobalConfigValidatorException(
                'meta.defaultView == "configuration" but there is no configuration defined in globalConfig'
            )
        if default_view == "inputs" and not self._global_config.has_inputs():
            raise GlobalConfigValidatorException(
                'meta.defaultView == "inputs" but there is no inputs defined in globalConfig'
            )
        if default_view == "dashboard" and not self._global_config.has_dashboard():
            raise GlobalConfigValidatorException(
                'meta.defaultView == "dashboard" but there is no dashboard defined in globalConfig'
            )

    def _validate_custom_search_commands(self) -> None:
        for command in self._global_config.custom_search_commands:
            file_path = os.path.join(self._source_dir, "bin", command["fileName"])
            if not os.path.isfile(file_path):
                raise GlobalConfigValidatorException(
                    f"{command['fileName']} is not present in `{os.path.join(self._source_dir, 'bin')}` directory. "
                    "Please ensure the file exists."
                )

            if (command.get("requiredSearchAssistant", False) is False) and (
                command.get("description")
                or command.get("usage")
                or command.get("syntax")
            ):
                logger.warning(
                    "requiredSearchAssistant is set to false "
                    "but attributes required for 'searchbnf.conf' is defined which is not required."
                )
            if (command.get("requiredSearchAssistant", False) is True) and not (
                command.get("description")
                and command.get("usage")
                and command.get("syntax")
            ):
                raise GlobalConfigValidatorException(
                    "One of the attributes among `description`, `usage`, `syntax`"
                    " is not been defined in globalConfig. Define them as requiredSearchAssistant is set to True."
                )

            if command["commandName"] in SPLUNK_COMMANDS:
                raise GlobalConfigValidatorException(
                    f"CommandName: {command['commandName']}"
                    " cannot have the same name as Splunk built-in command."
                )

            fileName_without_extension = command["fileName"].replace(".py", "")
            if command["commandName"] == fileName_without_extension:
                # Here we are generating file based on commandName therefore
                # the core logic should not have the same name as commandName
                raise GlobalConfigValidatorException(
                    f"Filename: {fileName_without_extension} and CommandName: {command['commandName']}"
                    " should not be same for custom search command."
                )

    def _validate_if_entities_has_oauth_configured_correctly(
        self, tabs: list[dict[str, Any]]
    ) -> None:
        """
        Validates if entities has oauth configured correctly.
        """
        grouped_entities: list[Any] = [
            el.get("entity") for el in tabs if el.get("entity")
        ]
        all_entities = list(itertools.chain.from_iterable(grouped_entities))

        for entity in all_entities:
            if entity["type"] == "oauth":
                # each auth_type defined in list should have entities defined
                for auth_type in entity["options"]["auth_type"]:
                    if auth_type not in entity["options"]:
                        raise GlobalConfigValidatorException(
                            f"Authorization type '{auth_type}' does not have any entities defined."
                        )

                if "oauth_type_labels" in entity["options"]:
                    for auth_type_in_label in entity["options"][
                        "oauth_type_labels"
                    ].keys():
                        if auth_type_in_label not in entity["options"]:
                            raise GlobalConfigValidatorException(
                                f"Authorization type '{auth_type_in_label}', included in "
                                "oauth_type_labels, does not have any entities defined."
                            )

    def _validate_oauth_entities_definition(self) -> None:
        """
        Validates that OAuth defined in oauth.options.auth_type
        has definitions for entities under options["auth_type"]
        """
        pages = self._config["pages"]

        if "configuration" in pages:
            # tabs are required in configuration
            tabs = pages["configuration"]["tabs"]

            self._validate_if_entities_has_oauth_configured_correctly(tabs)

    def _validate_usage_of_placeholder(self) -> None:
        """
        Stops the build of addon and logs error if placeholder is used.
        Deprecation Notice: https://github.com/splunk/addonfactory-ucc-generator/issues/831.
        """
        log_msg = (
            "`placeholder` option found for %s '%s' -> entity field '%s'. "
            "We recommend to use `help` instead (https://splunk.github.io/addonfactory-ucc-generator/entity/)."
            "\n\tDeprecation notice: https://github.com/splunk/addonfactory-ucc-generator/issues/831."
        )
        exc_msg = (
            "`placeholder` option found for %s '%s'. It has been removed from UCC. "
            "We recommend to use `help` instead (https://splunk.github.io/addonfactory-ucc-generator/entity/)."
        )
        for tab in self._global_config.configuration:
            for entity in tab.get("entity", []):
                if "placeholder" in entity.get("options", {}):
                    logger.error(
                        log_msg % ("configuration tab", tab["name"], entity["field"])
                    )
                    raise GlobalConfigValidatorException(
                        exc_msg % ("configuration tab", tab["name"])
                    )
        services = self._global_config.inputs
        if not services:
            return
        for service in services:
            for entity in service.get("entity", {}):
                if "placeholder" in entity.get("options", {}):
                    logger.error(
                        log_msg % ("input service", service["name"], entity["field"])
                    )
                    raise GlobalConfigValidatorException(
                        exc_msg % ("input service", service["name"])
                    )

    def _validate_user_defined_rest_handler_file(self) -> None:
        if self._global_config.content.get("options"):
            for restHandler in self._global_config.content["options"]["restHandlers"]:
                if restHandler.get("registerHandler"):
                    file_path = os.path.join(
                        self._source_dir, "bin", restHandler["registerHandler"]["file"]
                    )
                    if not os.path.isfile(file_path):
                        raise GlobalConfigValidatorException(
                            f"{restHandler['registerHandler']['file']} is not present in "
                            f"`{os.path.join(self._source_dir, 'bin')}` directory. Please ensure the file exists."
                        )

    def validate(self) -> None:
        self._validate_config_against_schema()
        if self._global_config.has_pages():
            self._validate_configuration_tab_table_has_name_field()
            self._validate_custom_rest_handlers()
            self._validate_file_type_entity()
            self._validate_validators()
            self._validate_multilevel_menu()
            self._validate_duplicates()
            self._validate_panels()
            self._validate_checkbox_group()
            self._validate_groups()
            self._validate_oauth_entities_definition()
            self._validate_field_modifications()
            self._validate_custom_search_commands()
            self._validate_usage_of_placeholder()
        self._validate_alerts()
        self._validate_meta_default_view()
        self._validate_user_defined_rest_handler_file()


def should_warn_on_empty_validators(entity: dict[str, Any]) -> bool:
    entity_type = entity.get("type")

    if entity_type in ENTITY_TYPES_WITHOUT_VALIDATORS:
        return False

    # special cases
    if entity_type == "oauth":
        return _should_warn_on_empty_validators_oauth(entity)

    elif entity_type == "checkboxGroup":
        return _should_warn_on_empty_validators_checkbox_group(entity)

    return "validators" not in entity


def _should_warn_on_empty_validators_checkbox_group(entity: dict[str, Any]) -> bool:
    for row in entity.get("options", {}).get("rows", []):
        row_validators = row.get("input", {}).get("validators")

        if not row_validators:
            return True

    return "validators" not in entity


def _should_warn_on_empty_validators_oauth(entity: dict[str, Any]) -> bool:
    options = entity.get("options", {})

    for auth_type in ("basic", "oauth"):
        for oauth_field in options.get(auth_type, []):
            if "validators" not in oauth_field:
                return True

    return False



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations
import logging
import os
import sys
import re

import shutil

from splunk_add_on_ucc_framework import utils

logger = logging.getLogger("ucc_gen")

ADDON_NAME_RE_STR = (
    r"^(?!^(CON|PRN|AUX|NUL|COM[1-9]|LPT[1-9])$)"
    r"(?!.*\.(tar(\.gz)?|tgz|spl)$)"
    r"[A-Za-z_.-][A-Za-z0-9_.-]*[A-Za-z0-9_-]$"
)
ADDON_NAME_RE = re.compile(ADDON_NAME_RE_STR)
ADDON_REST_ROOT_RE_STR = r"^[\w-]+$"
ADDON_REST_ROOT_RE = re.compile(ADDON_REST_ROOT_RE_STR)
ADDON_INPUT_NAME_RE_STR = r"^[0-9a-zA-Z][\w-]*$"
ADDON_INPUT_NAME_RE = re.compile(ADDON_INPUT_NAME_RE_STR)

UCC_DOC_SITE = "https://splunk.github.io/addonfactory-ucc-generator"


def _is_valid(pattern: re.Pattern[str], string: str) -> bool:
    result = pattern.search(string)
    if result is None:
        return False
    return True


def _is_valid_addon_name(addon_name: str) -> bool:
    return _is_valid(ADDON_NAME_RE, addon_name)


def _is_valid_rest_root_name(rest_root: str) -> bool:
    return _is_valid(ADDON_REST_ROOT_RE, rest_root)


def _is_valid_input_name(input_name: str) -> bool:
    if len(input_name) > 50:
        return False
    return _is_valid(ADDON_INPUT_NAME_RE, input_name)


def _generate_addon(
    addon_name: str,
    addon_display_name: str,
    addon_input_name: str,
    addon_version: str,
    addon_rest_root: str | None = None,
    overwrite: bool = False,
    need_proxy: bool = False,
    add_license: str | None = None,
    include_author: str | None = None,
) -> str:
    generated_addon_path = os.path.join(
        os.getcwd(),
        addon_name,
    )
    if not overwrite and os.path.exists(generated_addon_path):
        logger.error(
            "The location is already taken, use `--overwrite` "
            "option to overwrite the content of existing folder."
        )
        sys.exit(1)
    if overwrite:
        try:
            shutil.rmtree(generated_addon_path)
        except FileNotFoundError:
            pass
    os.makedirs(generated_addon_path)
    readme_path = os.path.join(generated_addon_path, "README.md")
    with open(readme_path, "w") as _f:
        _f.write(f"# {addon_name}\n")
    global_config_path = os.path.join(generated_addon_path, "globalConfig.json")
    global_config_rendered_content = (
        utils.get_j2_env()
        .get_template("globalConfig.json.init-template")
        .render(
            addon_name=addon_name,
            addon_rest_root=addon_rest_root,
            addon_version=addon_version,
            addon_display_name=addon_display_name,
            addon_input_name=addon_input_name,
            need_proxy=need_proxy,
        )
    )
    with open(global_config_path, "w") as _f:
        _f.write(global_config_rendered_content)
    package_path = os.path.join(generated_addon_path, "package")
    os.makedirs(package_path)
    icons_path = os.path.join(package_path, "static")
    os.makedirs(icons_path)
    for name in [
        "appIcon_2x.png",
        "appIcon.png",
        "appIconAlt_2x.png",
        "appIconAlt.png",
    ]:
        icons = utils.get_icons_path(name)
        shutil.copy(icons, icons_path)
    package_license_dir = os.path.join(package_path, "LICENSES")
    os.makedirs(package_license_dir)

    if add_license:
        package_license_path = os.path.join(package_license_dir, f"{add_license}.txt")
        template_path = utils.get_license_path(add_license)
        with open(template_path) as content:
            license_content = content.read()
        with open(package_license_path, "w") as _f:
            _f.write(license_content)
    else:
        # add-on licenses are to be kept in package/LICENSES directory only
        package_license_path = os.path.join(package_license_dir, "LICENSE.txt")
        with open(package_license_path, "w") as _f:
            pass
    package_readme_path = os.path.join(package_path, "README.txt")
    with open(package_readme_path, "w") as _f:
        pass
    package_app_manifest_path = os.path.join(package_path, "app.manifest")
    app_manifest_rendered_content = (
        utils.get_j2_env()
        .get_template("app.manifest.init-template")
        .render(
            addon_name=addon_name,
            addon_version=addon_version,
            addon_display_name=addon_display_name,
            add_license=add_license,
            include_author=include_author,
        )
    )
    with open(package_app_manifest_path, "w") as _f:
        _f.write(app_manifest_rendered_content)
    package_bin_path = os.path.join(package_path, "bin")
    os.makedirs(package_bin_path)

    package_bin_helper_path = os.path.join(
        package_bin_path, f"{addon_input_name}_helper.py"
    )
    helper_rendered_content = (
        utils.get_j2_env()
        .get_template("input.helper-init-template")
        .render(
            addon_name=addon_name,
            addon_rest_root=addon_rest_root,
            addon_input_name=addon_input_name,
        )
    )
    with open(package_bin_helper_path, "w") as _f:
        _f.write(helper_rendered_content + "\n")

    package_lib_path = os.path.join(package_path, "lib")
    os.makedirs(package_lib_path)
    package_lib_requirements_path = os.path.join(package_lib_path, "requirements.txt")
    with open(package_lib_requirements_path, "w") as _f:
        _f.writelines(
            [
                "splunktaucclib\n",
                "splunk-sdk\n",
                "solnlib\n",
            ]
        )

    # Packages that will be excluded from installation
    package_lib_requirements_exclude_path = os.path.join(
        package_lib_path, "exclude.txt"
    )

    with open(package_lib_requirements_exclude_path, "w") as fp:
        fp.writelines(
            "# Exclude the following packages from being installed in the add-on\n"
            "# You can add here packages that are shipped with Splunk by default\n"
            "# Please add one dependency name per line, similarly to requirements.txt\n"
        )

    return generated_addon_path


def init(
    addon_name: str,
    addon_display_name: str,
    addon_input_name: str,
    addon_version: str,
    addon_rest_root: str | None = None,
    overwrite: bool = False,
    need_proxy: bool = False,
    add_license: str | None = None,
    include_author: str | None = None,
) -> str:
    if not _is_valid_addon_name(addon_name):
        logger.error(
            f"Add-on name provided is not valid, it should follow '{ADDON_NAME_RE_STR}' regex."
        )
        sys.exit(1)

    if addon_rest_root is None:
        if _is_valid_rest_root_name(addon_name):
            logger.info(
                "Parameter `--addon-rest-root` is not provided, using `--addon-name` as a REST root."
            )
            addon_rest_root = addon_name
        else:
            logger.error(
                f"Can not use add-on name provided as an add-on REST root. "
                f"It should follow '{ADDON_REST_ROOT_RE_STR}' regex. "
                f"Please provide `--addon-rest-root` to specify add-on REST root."
            )
            sys.exit(1)
    else:
        if not _is_valid_rest_root_name(addon_rest_root):
            logger.error(
                f"Add-on REST root provided is not valid, it should follow '{ADDON_REST_ROOT_RE_STR}' regex."
            )
            sys.exit(1)
    if not _is_valid_input_name(addon_input_name):
        logger.error(
            f"Add-on input name provided is not valid, "
            f"it should follow '{ADDON_INPUT_NAME_RE_STR}' regex and be less than 50 characters."
        )
        sys.exit(1)
    if include_author == "":
        logger.error("The author name cannot be left empty, please provide some input.")
        sys.exit(1)
    if include_author:
        include_author = include_author.strip()
    generated_addon_path = _generate_addon(
        addon_name,
        addon_display_name,
        addon_input_name,
        addon_version,
        addon_rest_root,
        overwrite,
        need_proxy,
        add_license,
        include_author,
    )
    logger.info(f"Generated add-on is located here {generated_addon_path}")
    if add_license:
        logger.info(f"Specified {add_license} has been added to Licenses directory.")
    logger.info("README.txt is empty, you may need to modify the content of this file.")
    logger.info(
        f"""You have generated your add-on and are wondering what to do next? \
UCC offers many solutions to improve add-ons!
    Be sure to check out the following functionalities first:
        * Want to know what's going on under the hood of your add-on? \
Add our \"Monitoring dashboard\". - {UCC_DOC_SITE}/dashboard/
        * The default behavior of the UCC-generated REST handler is insufficient \
and you would like to add your own logic? - \
{UCC_DOC_SITE}/advanced/custom_rest_handler/
        * Your add-on is supposed to work on many OSs but uses libraries that are not cross-platform compatible? - \
{UCC_DOC_SITE}/advanced/os-dependent_libraries/\n
    Additional ways to improve add-on's look and feel:
        * Need better organization in your add-on? You can use the option of dividing inputs into categories - \
{UCC_DOC_SITE}/inputs/multilevel_menu/
        * Would you like to modify the value of one field depending on the value of another? - \
{UCC_DOC_SITE}/entity/modifyFieldsOnValue/
        * Would you like to automate your add-on's setup? Use autogenerated openapi.json file - \
{UCC_DOC_SITE}/openapi/
        * Does your add-on need some custom UI components? Check what support UCC offers - \
{UCC_DOC_SITE}/custom_ui_extensions/overview/

    The full range of possibilities offered by UCC can be found in the official documentation - \
{UCC_DOC_SITE}

    To install this newly generated add-on on your Splunk instance, you need to follow the next two steps in this order:
        * Build the add-on you just generated using the \"build\" command - \
{UCC_DOC_SITE}/quickstart/#ucc-gen-build
        * Package the built add-on up using the \"package\" command so that you can install it on your own \
Splunk instance - {UCC_DOC_SITE}/quickstart/#ucc-gen-package"""
    )
    return generated_addon_path
